"""Forcing."""
import abc
import copy
import json
import logging
import os
import shutil
import time

import netCDF4
import numpy as np
import toml

from .cache import Cache
from .configuration import ConfigurationFromHarmonie
from .datetime_utils import as_datetime, as_timedelta
from .file import ForcingFileNetCDF
from .geo import get_geo_object
from .read import ConstantValue, ConvertedInput, Converter
from .util import deep_update


# TODO: should be abstract?
class SurfexForcing(object):
    """Surfex forcing class."""

    def __init__(self):
        """Construct object."""


class SurfexNetCDFForcing(SurfexForcing):
    """NetCDF surfex forcing."""

    def __init__(self, filename, geo):
        """Construct netcdf forcing."""
        SurfexForcing.__init__(self)
        self.io_object = ForcingFileNetCDF(filename, geo)


class SurfexOutputForcing(object):
    """Main output class for SURFEX forcing."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, base_time, geo, ntimes, var_objs, time_step_intervall):
        """Construct output forcing.

        Args:
            base_time (_type_): _description_
            geo (_type_): _description_
            ntimes (_type_): _description_
            var_objs (_type_): _description_
            time_step_intervall (_type_): _description_
        """
        self.time_step_intervall = time_step_intervall
        self.valid_time = None
        self.base_time = base_time
        self.geo = geo
        self.ntimes = ntimes
        self.time_step = 0
        self.time_step_value = 0
        self.var_objs = var_objs
        self._check_sanity()

    def _check_sanity(self):
        if len(self.var_objs) != self.nparameters:
            raise Exception(
                f"Inconsistent number of parameter. {str(len(self.var_objs))} != "
                f"{str(self.nparameters)}"
            )

        # Check if all parameters are present
        for var_obj in self.var_objs:
            self.parameters[var_obj.var_name] = 1

        for key, value in self.parameters.items():
            if value == 0:
                raise Exception(f"Required parameter {str(key)} is missing!")

    # Time dependent parameter
    nparameters = 11
    parameters = {
        "TA": 0,
        "QA": 0,
        "PS": 0,
        "DIR_SW": 0,
        "SCA_SW": 0,
        "LW": 0,
        "RAIN": 0,
        "SNOW": 0,
        "WIND": 0,
        "WIND_DIR": 0,
        "CO2": 0,
    }

    @abc.abstractmethod
    def write_forcing(self, var_objs, this_time, cache):
        """Write forcing."""
        raise NotImplementedError("users must define writeForcing to use this base class")


class NetCDFOutput(SurfexOutputForcing):
    """Forcing in NetCDF format."""

    forcing_file = {}

    translation = {
        "TA": "Tair",
        "QA": "Qair",
        "PS": "PSurf",
        "DIR_SW": "DIR_SWdown",
        "SCA_SW": "SCA_SWdown",
        "LW": "LWdown",
        "RAIN": "Rainf",
        "SNOW": "Snowf",
        "WIND": "Wind",
        "WIND_DIR": "Wind_DIR",
        "CO2": "CO2air",
    }

    def __init__(
        self,
        base_time,
        geo,
        fname,
        ntimes,
        var_objs,
        att_objs,
        att_time,
        cache,
        time_step,
        fmt="netcdf",
    ):
        """Construct netcdf forcing.

        Args:
            base_time (_type_): _description_
            geo (_type_): _description_
            fname (_type_): _description_
            ntimes (_type_): _description_
            var_objs (_type_): _description_
            att_objs (_type_): _description_
            att_time (_type_): _description_
            cache (_type_): _description_
            time_step (_type_): _description_
            fmt (str, optional): _description_. Defaults to "netcdf".

        Raises:
            NotImplementedError: NotImplementedError

        """
        SurfexOutputForcing.__init__(self, base_time, geo, ntimes, var_objs, time_step)
        if fmt == "netcdf":
            self.output_format = "NETCDF3_64BIT"
        elif fmt == "nc4":
            self.output_format = "NETCDF4"
        else:
            raise NotImplementedError(format)
        logging.info("Forcing type is netCDF")
        self.forcing_file = {}
        if fname is None:
            fname = "FORCING.nc"
        self.fname = fname
        self.tmp_fname = self.fname + ".tmp"
        self.file_handler = netCDF4.Dataset(
            self.tmp_fname, "w", format=self.output_format
        )
        self._define_forcing(geo, att_objs, att_time, cache)

    def write_forcing(self, var_objs, this_time, cache):
        """Write forcing.

        Args:
            var_objs (_type_): _description_
            this_time (_type_): _description_
            cache (_type_): _description_

        """
        # VARS
        for this_obj in self.var_objs:
            this_var = this_obj.var_name
            logging.info("Preparing variable %s", this_obj.var_name)
            tic = time.time()
            field = this_obj.read_time_step(this_time, cache)
            field = field.reshape([self.geo.nlats, self.geo.nlons], order="F").flatten()
            toc = time.time()
            logging.info("Preparation took %s seconds", str(toc - tic))
            self.forcing_file[self.translation[this_var]][self.time_step, :] = field

        self.forcing_file["TIME"][self.time_step] = self.time_step_value

    def _define_forcing(self, geo, att_objs, att_time, cache):
        logging.info("Define netcdf forcing")

        zs_oro = None
        zref = None
        uref = None
        for this_obj in att_objs:
            this_var = this_obj.var_name
            logging.info("Define: %s", this_obj.var_name)
            if this_var == "ZS":
                zs_oro = this_obj.read_time_step(att_time, cache)
                zs_oro = zs_oro.reshape(
                    [self.geo.nlats, self.geo.nlons], order="F"
                ).flatten()
            elif this_var == "ZREF":
                zref = this_obj.read_time_step(att_time, cache)
                zref = zref.reshape([self.geo.nlats, self.geo.nlons], order="F").flatten()
            elif this_var == "UREF":
                uref = this_obj.read_time_step(att_time, cache)
                uref = uref.reshape([self.geo.nlats, self.geo.nlons], order="F").flatten()

        # DIMS
        self.forcing_file["NPOINTS"] = self.file_handler.createDimension(
            "Number_of_points", geo.npoints
        )
        self.forcing_file["NTIMES"] = self.file_handler.createDimension(
            "time", self.ntimes
        )

        # DEFINE VARS
        self.forcing_file["TIME"] = self.file_handler.createVariable(
            "time", "f4", ("time",)
        )
        self.forcing_file["TIME"].units = (
            "hours since " + f"{self.base_time.strftime('%Y-%m-%d %H')}:00:00 0:00"
        )
        self.forcing_file["TSTEP"] = self.file_handler.createVariable(
            "FRC_TIME_STP", "f4"
        )
        self.forcing_file["TSTEP"].longname = "Forcing_Time_Step"
        self.forcing_file["TSTEP"][:] = self.time_step_intervall
        self.forcing_file["LON"] = self.file_handler.createVariable(
            "LON", "f4", ("Number_of_points",)
        )
        self.forcing_file["LON"].longname = "Longitude"
        self.forcing_file["LON"][:] = geo.lonlist
        self.forcing_file["LAT"] = self.file_handler.createVariable(
            "LAT", "f4", ("Number_of_points",)
        )
        self.forcing_file["LAT"].longname = "Latitude"
        self.forcing_file["LAT"][:] = geo.latlist
        self.forcing_file["ZS"] = self.file_handler.createVariable(
            "ZS", "f4", ("Number_of_points",)
        )
        self.forcing_file["ZS"].longname = "Surface_Orography"
        self.forcing_file["ZS"][:] = zs_oro
        self.forcing_file["ZREF"] = self.file_handler.createVariable(
            "ZREF", "f4", ("Number_of_points",)
        )
        self.forcing_file["ZREF"].longname = "Reference_Height"
        self.forcing_file["ZREF"].units = "m"
        self.forcing_file["ZREF"][:] = zref
        self.forcing_file["UREF"] = self.file_handler.createVariable(
            "UREF", "f4", ("Number_of_points",)
        )
        self.forcing_file["UREF"].longname = "Reference_Height_for_Wind"
        self.forcing_file["UREF"].units = "m"
        self.forcing_file["UREF"][:] = uref

        # Define time dependent variables
        for this_obj in self.var_objs:
            this_var = this_obj.var_name

            if this_var == "TA":
                self.forcing_file["Tair"] = self.file_handler.createVariable(
                    "Tair",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Tair"].longname = "Near_Surface_Air_Temperature"
                self.forcing_file["Tair"].units = "K"
            elif this_var == "QA":
                self.forcing_file["Qair"] = self.file_handler.createVariable(
                    "Qair",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Qair"].longname = "Near_Surface_Specific_Humidity"
                self.forcing_file["Qair"].units = "kg/kg"
            elif this_var == "PS":
                self.forcing_file["PSurf"] = self.file_handler.createVariable(
                    "PSurf",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["PSurf"].longname = "Surface_Pressure"
                self.forcing_file["PSurf"].units = "Pa"
            elif this_var == "DIR_SW":
                self.forcing_file["DIR_SWdown"] = self.file_handler.createVariable(
                    "DIR_SWdown",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file[
                    "DIR_SWdown"
                ].longname = "Surface_Incident_Downwelling_Shortwave_Radiation"
                self.forcing_file["DIR_SWdown"].units = "W/m2"
            elif this_var == "SCA_SW":
                self.forcing_file["SCA_SWdown"] = self.file_handler.createVariable(
                    "SCA_SWdown",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file[
                    "SCA_SWdown"
                ].longname = "Surface_Incident_Diffuse_Shortwave_Radiation"
                self.forcing_file["SCA_SWdown"].units = "W/m2"
            elif this_var == "LW":
                self.forcing_file["LWdown"] = self.file_handler.createVariable(
                    "LWdown",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file[
                    "LWdown"
                ].longname = "Surface_Incident_Diffuse_Longwave_Radiation"
                self.forcing_file["LWdown"].units = "W/m2"
            elif this_var == "RAIN":
                self.forcing_file["Rainf"] = self.file_handler.createVariable(
                    "Rainf",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Rainf"].longname = "Rainfall_Rate"
                self.forcing_file["Rainf"].units = "kg/m2/s"
            elif this_var == "SNOW":
                self.forcing_file["Snowf"] = self.file_handler.createVariable(
                    "Snowf",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Snowf"].longname = "Snowfall_Rate"
                self.forcing_file["Snowf"].units = "kg/m2/s"
            elif this_var == "WIND":
                self.forcing_file["Wind"] = self.file_handler.createVariable(
                    "Wind",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Wind"].longname = "Wind_Speed"
                self.forcing_file["Wind"].units = "m/s"
            elif this_var == "WIND_DIR":
                self.forcing_file["Wind_DIR"] = self.file_handler.createVariable(
                    "Wind_DIR",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Wind_DIR"].longname = "Wind_Direction"
            elif this_var == "CO2":
                self.forcing_file["CO2air"] = self.file_handler.createVariable(
                    "CO2air",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["CO2air"].longname = "Near_Surface_CO2_Concentration"
                self.forcing_file["CO2air"].units = "kg/m3"
            else:
                raise NotImplementedError(
                    f"This should never happen! {this_var} is not defined!"
                )

    def finalize(self):
        """Finalize the forcing. Close the file."""
        logging.debug("Close file")
        self.file_handler.close()
        shutil.move(self.tmp_fname, self.fname)


class AsciiOutput(SurfexOutputForcing):
    """Forcing in ASCII format."""

    def __init__(
        self,
        base_time,
        geo,
        fname,
        ntimes,
        var_objs,
        att_objs,
        att_time,
        cache,
        time_step,
    ):
        """Construct ASCII forcing output."""
        SurfexOutputForcing.__init__(self, base_time, geo, ntimes, var_objs, time_step)
        self.output_format = "ascii"
        logging.info("Forcing type is ASCII")
        self.forcing_file = {}
        self.file_handler = {}
        if fname is None:
            fname = "Params_config.txt"
        self.fname = fname
        self._define_forcing(geo, att_objs, att_time, cache)

    def write_forcing(self, var_objs, this_time, cache):
        """Write forcing.

        Args:
            var_objs (_type_): _description_
            this_time (_type_): _description_
            cache (_type_): _description_

        """
        for this_obj in self.var_objs:
            this_var = this_obj.var_name
            logging.info("Write var name %s", this_obj.var_name)
            field = this_obj.read_time_step(this_time, cache)
            fmt = "%20.8f"
            cols = 50
            write_formatted_array(self.file_handler[this_var], field, cols, fmt)

    def _define_forcing(self, geo, att_objs, att_time, cache):
        zs_oro = None
        zref = None
        uref = None
        for this_obj in att_objs:
            this_var = this_obj.var_name
            if this_var == "ZS":
                zs_oro = this_obj.read_time_step(att_time, cache)
            elif this_var == "ZREF":
                zref = this_obj.read_time_step(att_time, cache)
            elif this_var == "UREF":
                uref = this_obj.read_time_step(att_time, cache)

        second = (
            self.base_time
            - self.base_time.replace(hour=0, minute=0, second=0, microsecond=0)
        ).total_seconds()
        fmt = "%15.8f"
        cols = 50
        file_handler = open(self.fname, "w", encoding="utf-8")
        file_handler.write(str(geo.npoints) + "\n")
        file_handler.write(str(self.ntimes) + "\n")
        file_handler.write(str(self.time_step_intervall) + "\n")
        file_handler.write(self.base_time.strftime("%Y") + "\n")
        file_handler.write(self.base_time.strftime("%m") + "\n")
        file_handler.write(self.base_time.strftime("%d") + "\n")
        file_handler.write(str(second) + "\n")
        write_formatted_array(file_handler, geo.lons, cols, fmt)
        write_formatted_array(file_handler, geo.lats, cols, fmt)
        write_formatted_array(file_handler, zs_oro, cols, fmt)
        write_formatted_array(file_handler, zref, cols, fmt)
        write_formatted_array(file_handler, uref, cols, fmt)
        file_handler.close()

        for key in self.parameters:
            nam = key
            if key == "WIND_DIR":
                nam = "DIR"
            self.forcing_file[key] = "Forc_" + nam + ".txt"
            self.file_handler[key] = open(self.forcing_file[key], "w", encoding="utf-8")

    def finalize(self):
        """Finalize forcing."""
        logging.debug("Close file")
        for key in self.parameters:
            self.file_handler[key].close()


def write_formatted_array(file, array, columns, fileformat):
    """Write a formatted array."""
    astr = np.empty(array.size - array.size % columns, dtype="float64")
    astr = array[0 : astr.size]
    astr = astr.reshape((columns, astr.size / columns), order="F")
    mlw = (len(fileformat % 0)) * (columns + 1)
    formatter = {"float_kind": lambda x: fileformat % x}
    astr_end = np.array2string(
        array[astr.size :], separator="", max_line_width=mlw, formatter=formatter
    )[1:-1]
    np.savetxt(file, astr.T, fmt=fileformat, newline="\n", delimiter="")
    file.write(astr_end + "\n")


def run_time_loop(options, var_objs, att_objs):
    """Run time loop."""
    tic = time.time()
    this_time = options["start"]

    cache = Cache(options["cache_interval"])
    time_step = options["timestep"]
    single = False
    if "single" in options:
        single = options["single"]

    # Find how many time steps we want to write
    ntimes = 0
    while this_time <= options["stop"]:
        ntimes = ntimes + 1
        this_time = this_time + as_timedelta(seconds=options["timestep"])
    if single:
        time_step = 1
        if ntimes == 1:
            ntimes = 2
            logging.info("Print single time step twice %s", 0)
        else:
            raise Exception("Option single should be used with one time step")

    # Create output object
    if (
        str.lower(options["output_format"]) == "netcdf"
        or str.lower(options["output_format"]) == "nc4"
    ):
        # Set att_time the same as start
        att_time = options["start"]
        output = NetCDFOutput(
            options["start"],
            options["geo_out"],
            options["output_file"],
            ntimes,
            var_objs,
            att_objs,
            att_time,
            cache,
            time_step,
            fmt=str.lower(options["output_format"]),
        )
    elif str.lower(options["output_format"]) == "ascii":
        att_time = options["start"]
        output = AsciiOutput(
            options["start"],
            options["geo_out"],
            options["output_file"],
            ntimes,
            var_objs,
            att_objs,
            att_time,
            cache,
            time_step,
        )
    else:
        raise NotImplementedError("Invalid output format " + options["output_format"])

    # Loop output time steps
    this_time = options["start"]
    while this_time <= options["stop"]:

        # Write for each time step
        logging.info(
            "Creating forcing for: %s  time_step: %s",
            this_time.strftime("%Y%m%d%H"),
            str(output.time_step),
        )
        output.write_forcing(var_objs, this_time, cache)
        output.time_step = output.time_step + 1
        if not single:
            output.time_step_value = output.time_step
            this_time = this_time + as_timedelta(seconds=options["timestep"])
            if cache is not None:
                cache.clean_fields(this_time)
        else:
            output.time_step_value = 0
            if output.time_step > 1:
                this_time = this_time + as_timedelta(seconds=options["timestep"])

    # Finalize forcing
    output.finalize()
    toc = time.time()
    logging.info("Forcing generation took %s seconds", str(toc - tic))


def set_input_object(
    sfx_var,
    merged_conf,
    geo,
    forcingformat,
    selected_converter,
    ref_height,
    first_base_time,
    timestep,
):
    """Set the input parameter for a specific SURFEX forcing variable based on input.

    Args:
        sfx_var (str): _description_
        merged_conf (_type_): _description_
        geo (_type_): _description_
        forcingformat (_type_): _description_
        selected_converter (_type_): _description_
        ref_height (_type_): _description_
        first_base_time (_type_): _description_
        timestep (_type_): _description_

    Returns:
        _type_: _description_

    Raises:
        KeyError: KeyError

    """
    #########################################
    # 1. Gobal configuration from yaml file
    #########################################

    conf = copy.deepcopy(merged_conf)

    # Now we know the CLA settings for each surfex variable

    # Set defaults (merged global and user settings)
    # Theses must be sent to the converter to be used for each variable
    defs = {}
    if forcingformat in conf:
        defs = copy.deepcopy(conf[forcingformat])
        defs.update({"timestep": timestep})

    # All objects with converters, find converter dict entry
    conf_dict = {}
    if forcingformat != "constant":

        # Non-height dependent variables
        if ref_height is None:
            if "converter" in conf[sfx_var][forcingformat]:
                conf_dict = copy.deepcopy(conf[sfx_var][forcingformat]["converter"])
            else:
                raise KeyError("No converter defined for " + sfx_var)

        # Variables with height dependency
        else:
            if ref_height in conf[sfx_var]:
                if forcingformat not in conf[sfx_var][ref_height]:
                    msg = (
                        f"{str(conf[sfx_var])}: "
                        + f"Missing definitions for {sfx_var} and format: {forcingformat}"
                    )
                    raise KeyError(msg)
                if conf[sfx_var][ref_height][forcingformat] is None:
                    raise KeyError(
                        f"{str(conf[sfx_var])}: Missing definitions for {sfx_var}"
                    )
                if "converter" in conf[sfx_var][ref_height][forcingformat]:
                    conf_dict = copy.deepcopy(
                        conf[sfx_var][ref_height][forcingformat]["converter"]
                    )
                else:
                    raise KeyError("No converter defined for " + sfx_var)
            else:
                raise KeyError(
                    'No ref height "' + ref_height + '" defined for ' + sfx_var
                )

    ##############################################################
    ##############################################################
    ##############################################################
    # Create the object to be returned
    if forcingformat == "constant":
        if ref_height is None:
            if "constant" in conf[sfx_var]:
                const_dict = copy.deepcopy(conf[sfx_var]["constant"])
            else:
                raise KeyError("No constant defined for " + sfx_var)
        else:
            if ref_height in conf[sfx_var]:
                if "constant" in conf[sfx_var][ref_height]:
                    const_dict = copy.deepcopy(conf[sfx_var][ref_height]["constant"])
                else:
                    raise KeyError("No constant defined for " + sfx_var)
            else:
                raise KeyError(
                    'No ref height "' + ref_height + '" defined for ' + sfx_var
                )

        obj = ConstantValue(geo, sfx_var, const_dict)
    else:

        # Construct the converter
        converter = Converter(
            selected_converter, first_base_time, defs, conf_dict, forcingformat
        )

        # Construct the input object
        obj = ConvertedInput(geo, sfx_var, converter)
    return obj


def set_forcing_config(**kwargs):
    """Set the forcing config."""
    file_base = None
    if "harmonie" in kwargs and kwargs["harmonie"]:
        config_exp = None
        if "config_exp_surfex" in kwargs:
            if kwargs["config_exp_surfex"] is not None:
                config_exp = kwargs["config_exp_surfex"]
        if config_exp is None:
            config_exp = (
                f"{os.path.abspath(os.path.dirname(__file__))}/cfg/config_exp_surfex.toml"
            )
        logging.info("Using default config from: %s", config_exp)
        input_data = toml.load(open(config_exp, "r", encoding="utf-8"))
        config = ConfigurationFromHarmonie(os.environ, input_data)
        geo_out = config.geo
    elif "domain" in kwargs and kwargs["domain"] is not None:
        geo_out = get_geo_object(json.load(open(kwargs["domain"], "r", encoding="utf-8")))
    else:
        raise Exception("No geometry is set")

    user_config = {}
    pattern = None
    timestep = 3600
    cache_interval = 3600
    zsoro = "default"
    zsoro_converter = "none"
    zval = "default"
    zval_converter = "none"
    uval = "default"
    uval_converter = "none"

    tair = "default"
    tair_converter = "none"
    qair = "default"
    qair_converter = "none"
    psurf = "default"
    psurf_converter = "none"
    dir_sw = "default"
    dir_sw_converter = "none"
    sca_sw = "default"
    sca_sw_converter = "none"
    lw_rad = "default"
    lw_rad_converter = "none"
    rain = "default"
    rain_converter = "none"
    snow = "default"
    snow_converter = "none"
    wind = "default"
    wind_converter = "none"
    wind_dir = "default"
    wind_dir_converter = "none"
    co2 = "default"
    co2_converter = "none"

    analysis = False
    interpolation = None
    try:
        dtg_start = kwargs["dtg_start"]
        dtg_stop = kwargs["dtg_stop"]
        input_format = kwargs["input_format"]
        output_format = kwargs["output_format"]
        outfile = kwargs["of"]
        zref = kwargs["zref"]
        uref = kwargs["uref"]
        config = kwargs["config"]
        if "interpolation" in kwargs:
            interpolation = kwargs["interpolation"]
        if "analysis" in kwargs:
            analysis = kwargs["analysis"]
        if "fb" in kwargs:
            file_base = kwargs["fb"]
        if "geo_out" in kwargs:
            geo_out = kwargs["geo_out"]
        if "user_config" in kwargs:
            user_config = kwargs["user_config"]
        if "pattern" in kwargs:
            pattern = kwargs["pattern"]
        if "timestep" in kwargs:
            timestep = kwargs["timestep"]
        if "cache_interval" in kwargs:
            cache_interval = kwargs["cache_interval"]
        if "zsoro" in kwargs:
            zsoro = kwargs["zsoro"]
        if "zsoro_converter" in kwargs:
            zsoro_converter = kwargs["zsoro_converter"]
        if "zval" in kwargs:
            zval = kwargs["zval"]
        if "zval_converter" in kwargs:
            zval_converter = kwargs["zval_converter"]
        if "uval_converter" in kwargs:
            uval_converter = kwargs["uval_converter"]
        if "uval" in kwargs:
            uval = kwargs["uval"]
        if "ta" in kwargs:
            tair = kwargs["ta"]
        if "ta_converter" in kwargs:
            tair_converter = kwargs["ta_converter"]
        if "qa" in kwargs:
            qair = kwargs["qa"]
        if "qa_converter" in kwargs:
            qair_converter = kwargs["qa_converter"]
        if "ps" in kwargs:
            psurf = kwargs["ps"]
        if "ps_converter" in kwargs:
            psurf_converter = kwargs["ps_converter"]
        if "dir_sw" in kwargs:
            dir_sw = kwargs["dir_sw"]
        if "dir_sw_converter" in kwargs:
            dir_sw_converter = kwargs["dir_sw_converter"]
        if "sca_sw" in kwargs:
            sca_sw = kwargs["sca_sw"]
        if "sca_sw_converter" in kwargs:
            sca_sw_converter = kwargs["sca_sw_converter"]
        if "lw" in kwargs:
            lw_rad = kwargs["lw"]
        if "lw_converter" in kwargs:
            lw_rad_converter = kwargs["lw_converter"]
        if "rain" in kwargs:
            rain = kwargs["rain"]
        if "rain_converter" in kwargs:
            rain_converter = kwargs["rain_converter"]
        if "snow" in kwargs:
            snow = kwargs["snow"]
        if "snow_converter" in kwargs:
            snow_converter = kwargs["snow_converter"]
        if "wind" in kwargs:
            wind = kwargs["wind"]
        if "wind_converter" in kwargs:
            wind_converter = kwargs["wind_converter"]
        if "wind_dir" in kwargs:
            wind_dir = kwargs["wind_dir"]
        if "wind_dir_converter" in kwargs:
            wind_dir_converter = kwargs["wind_dir_converter"]
        if "co2" in kwargs:
            co2 = kwargs["co2"]
        if "co2_converter" in kwargs:
            co2_converter = kwargs["co2_converter"]

    except ValueError:
        raise Exception("Needed input is missing") from ValueError

    # Time information
    if (int(dtg_start) or int(dtg_stop)) < 1000010100:
        raise Exception(
            "Invalid start and stop times! " + str(dtg_start) + " " + str(dtg_stop)
        )

    start = as_datetime(str.strip(str(dtg_start)))
    stop = as_datetime(str.strip(str(dtg_stop)))
    if file_base is None:
        first_base_time = start
    else:
        first_base_time = as_datetime(str.strip(str(file_base)))

    # Merge all settings with user all settings
    merged_conf = deep_update(config, user_config)

    # Replace global settings from
    fileformat = input_format
    if pattern is not None:
        merged_conf[fileformat]["filepattern"] = pattern

    # Interpolation
    merged_conf[fileformat]["interpolation"] = interpolation

    # Prefer forecast or analysis
    if analysis:
        merged_conf[fileformat]["prefer_forecast"] = False
        merged_conf[fileformat]["fcint"] = 3600.0
        merged_conf[fileformat]["offset"] = 0

    if "geo_input" in kwargs:
        geo_input = kwargs["geo_input"]
        if geo_input is not None:
            if os.path.exists(geo_input):
                geo_input = get_geo_object(
                    json.load(open(geo_input, "r", encoding="utf-8"))
                )
                merged_conf[fileformat]["geo_input"] = geo_input
            else:
                logging.info("Input geometry %s does not exist", geo_input)

    # Set attributes
    atts = ["ZS", "ZREF", "UREF"]
    att_objs = []
    for att_var in atts:

        # Override with command line options for a given variable
        ref_height = None
        cformat = fileformat
        if att_var == "ZS":
            if zsoro != "default":
                cformat = zsoro
            selected_converter = zsoro_converter
        elif att_var == "ZREF":
            if zval != "default":
                cformat = zval
            selected_converter = zval_converter
            ref_height = zref
            if ref_height == "screen":
                cformat = "constant"
        elif att_var == "UREF":
            if uval != "default":
                cformat = uval
            selected_converter = uval_converter
            ref_height = uref
            if ref_height == "screen":
                cformat = "constant"
        else:
            raise NotImplementedError

        att_objs.append(
            set_input_object(
                att_var,
                merged_conf,
                geo_out,
                cformat,
                selected_converter,
                ref_height,
                first_base_time,
                timestep,
            )
        )

    # Set forcing variables (time dependent)
    variables = [
        "TA",
        "QA",
        "PS",
        "DIR_SW",
        "SCA_SW",
        "LW",
        "RAIN",
        "SNOW",
        "WIND",
        "WIND_DIR",
        "CO2",
    ]
    var_objs = []
    # Search in config file for parameters to override
    for sfx_var in variables:

        ref_height = None
        cformat = fileformat
        if sfx_var == "TA":
            if tair != "default":
                cformat = tair
            selected_converter = tair_converter
            ref_height = zref
        elif sfx_var == "QA":
            if qair != "default":
                cformat = qair
            selected_converter = qair_converter
            ref_height = zref
        elif sfx_var == "PS":
            if psurf != "default":
                cformat = psurf
            selected_converter = psurf_converter
        elif sfx_var == "DIR_SW":
            if dir_sw != "default":
                cformat = dir_sw
            selected_converter = dir_sw_converter
        elif sfx_var == "SCA_SW":
            if sca_sw != "default":
                cformat = sca_sw
            selected_converter = sca_sw_converter
        elif sfx_var == "LW":
            if lw_rad != "default":
                cformat = lw_rad
            selected_converter = lw_rad_converter
        elif sfx_var == "RAIN":
            if rain != "default":
                cformat = rain
            selected_converter = rain_converter
        elif sfx_var == "SNOW":
            if snow != "default":
                cformat = snow
            selected_converter = snow_converter
        elif sfx_var == "WIND":
            if wind != "default":
                cformat = wind
            selected_converter = wind_converter
            ref_height = uref
        elif sfx_var == "WIND_DIR":
            if wind_dir != "default":
                cformat = wind_dir
            selected_converter = wind_dir_converter
            ref_height = uref
        elif sfx_var == "CO2":
            if co2 != "default":
                cformat = co2
            selected_converter = co2_converter
        else:
            raise NotImplementedError
        var_objs.append(
            set_input_object(
                sfx_var,
                merged_conf,
                geo_out,
                cformat,
                selected_converter,
                ref_height,
                first_base_time,
                timestep,
            )
        )

    # Save options
    options = dict()
    options["output_format"] = output_format
    options["output_file"] = outfile
    options["start"] = start
    options["stop"] = stop
    options["timestep"] = timestep
    options["geo_out"] = geo_out
    options["single"] = False
    if "single" in kwargs:
        options["single"] = kwargs["single"]
    options["cache_interval"] = cache_interval

    return options, var_objs, att_objs


def modify_forcing(**kwargs):
    """Modify forcing."""
    infile = kwargs["input_file"]
    outfile = kwargs["output_file"]
    time_step = kwargs["time_step"]
    variables = kwargs["variables"]

    ifile = netCDF4.Dataset(infile, "r")
    ofile = netCDF4.Dataset(outfile, "r+")

    for var in variables:
        print("Modify variable " + var)
        print(
            "input", ifile[var][time_step, :], ifile[var][time_step, :].shape, time_step
        )
        print("output", ofile[var][0, :], ofile[var][0, :].shape)
        ofile[var][0, :] = ifile[var][time_step, :]
        ofile.sync()

    ifile.close()
    ofile.close()



"""Namelist."""
import json
import logging
import os

import f90nml
import yaml

from .datetime_utils import as_datetime, as_timedelta
from .ecoclimap import Ecoclimap, EcoclimapSG


class BaseNamelist(object):
    """Base namelist."""

    def __init__(
        self,
        program,
        config,
        input_path,
        forc_zs=False,
        prep_file=None,
        prep_filetype=None,
        prep_pgdfile=None,
        prep_pgdfiletype=None,
        dtg=None,
        fcint=3,
        geo=None,
    ):
        """Construct a base namelists class to be implemented by namelist implementations.

        Args:
            program (str): Which surfex binary you want to run ["pgd", "prep", "offline", "soda"]
            config (surfex.Configuration): A SURFEX configuration object
            input_path (str): A path to search for json namelist blocks
            forc_zs (bool): Set the surfex orography heigth to the same as in the forcing files
            prep_file (str): Input file for prep
            prep_filetype (str): The format of prep_file
            prep_pgdfile (str): Input PGD file for prep in case it comes from SURFEX
            prep_pgdfiletype (str): The format of the prep_pgdfile
            dtg (datetime.datetime): The date/time you want to run
            fcint (int): The intervall between the cycles. Used for first guesses.
            geo (surfex.Geo): Surfex geometry. The domain you want to run on

        Raises:
            RuntimeError: Needed input
            NotImplementedError: Mode not implemented

        """
        self.config = config
        self.input_path = input_path
        self.forc_zs = forc_zs
        if dtg is not None:
            if isinstance(dtg, str):
                dtg = as_datetime(dtg)
        self.dtg = dtg
        check_parsing = True
        if self.dtg is None:
            check_parsing = False
        self.fcint = fcint
        self.geo = geo

        # The time stamp of next cycle file
        forecast_length = self.fcint
        if self.dtg is not None:
            self.end_of_forecast = self.dtg + as_timedelta(seconds=forecast_length * 3600)
        else:
            self.end_of_forecast = None

        logging.info("Creating JSON namelist input for program: %s", program)
        self.input_list = []

        self.prolog(check_parsing)
        # Program specific settings
        if program == "pgd":
            self.set_pgd_namelist()
        elif program == "prep":
            if prep_file is None:
                raise RuntimeError(
                    "Prep need an input file either as a json namelist or a surfex "
                    "supported format"
                )
            self.set_prep_namelist(
                prep_file=prep_file,
                prep_filetype=prep_filetype,
                prep_pgdfile=prep_pgdfile,
                prep_pgdfiletype=prep_pgdfiletype,
            )
        elif program == "offline" or program == "perturbed":
            self.set_offline_namelist()
        elif program == "soda":
            self.set_soda_namelist()
        else:
            raise NotImplementedError(program)
        self.epilog()
        self.override()

    def prolog(self, check_parsing):
        """Prolog.

        Args:
            check_parsing (_type_): _description_

        """
        # IO
        self.input_list.append({"file": self.input_path + "/io.json"})
        self.input_list.append(
            {
                "json": {
                    "NAM_IO_OFFLINE": {
                        "CSURF_FILETYPE": self.config.get_setting(
                            "SURFEX#IO#CSURF_FILETYPE"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_IO_OFFLINE": {
                        "CTIMESERIES_FILETYPE": self.config.get_setting(
                            "SURFEX#IO#CTIMESERIES_FILETYPE"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_IO_OFFLINE": {
                        "CFORCING_FILETYPE": self.config.get_setting(
                            "SURFEX#IO#CFORCING_FILETYPE"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_IO_OFFLINE": {
                        "CPGDFILE": self.config.get_setting("SURFEX#IO#CPGDFILE")
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_IO_OFFLINE": {
                        "CPREPFILE": self.config.get_setting("SURFEX#IO#CPREPFILE")
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_IO_OFFLINE": {
                        "CSURFFILE": self.config.get_setting(
                            "SURFEX#IO#CSURFFILE",
                            validtime=self.end_of_forecast,
                            basedtg=self.dtg,
                            check_parsing=check_parsing,
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_IO_OFFLINE": {
                        "XTSTEP_SURF": self.config.get_setting("SURFEX#IO#XTSTEP")
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_IO_OFFLINE": {
                        "XTSTEP_OUTPUT": self.config.get_setting(
                            "SURFEX#IO#XTSTEP_OUTPUT"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_WRITE_SURF_ATM": {
                        "LSPLIT_PATCH": self.config.get_setting("SURFEX#IO#LSPLIT_PATCH")
                    }
                }
            }
        )

        if self.forc_zs:
            self.input_list.append({"json": {"NAM_IO_OFFLINE": {"LSET_FORC_ZS": True}}})

        # Constants and parameters
        self.input_list.append({"file": self.input_path + "/constants.json"})
        self.input_list.append(
            {
                "json": {
                    "NAM_SURF_ATM": {
                        "XRIMAX": self.config.get_setting("SURFEX#PARAMETERS#XRIMAX")
                    }
                }
            }
        )

    def set_pgd_namelist(self):
        """Set pgd namelist."""
        # PGS schemes
        self.input_list.append(
            {
                "json": {
                    "NAM_PGD_SCHEMES": {
                        "CSEA": self.config.get_setting("SURFEX#TILES#SEA"),
                        "CWATER": self.config.get_setting("SURFEX#TILES#INLAND_WATER"),
                        "CNATURE": self.config.get_setting("SURFEX#TILES#NATURE"),
                        "CTOWN": self.config.get_setting("SURFEX#TILES#TOWN"),
                    }
                }
            }
        )

        eco_sg = self.config.get_setting("SURFEX#COVER#SG")
        # Ecoclimap SG
        self.input_list.append({"json": {"NAM_FRAC": {"LECOSG": eco_sg}}})
        if self.config.get_setting("SURFEX#COVER#SG"):
            ecoclimap = EcoclimapSG(self.config)

            self.input_list.append(
                {"json": {"NAM_DATA_ISBA": {"NTIME": ecoclimap.decades}}}
            )

            fname = self.config.get_setting("SURFEX#COVER#H_TREE")
            if fname != "" and fname is not None:
                self.input_list.append(
                    self.set_dirtyp_data_namelist(
                        "NAM_DATA_ISBA", "H_TREE", fname, vtype=1
                    )
                )

            decadal_data_types = [
                "ALBNIR_SOIL",
                "ALBNIR_VEG",
                "ALBVIS_SOIL",
                "ALBVIS_VEG",
                "LAI",
            ]
            for decadal_data_type in decadal_data_types:
                for vtt in range(1, ecoclimap.veg_types + 1):
                    for decade in range(1, ecoclimap.decades + 1):
                        filepattern = self.config.get_setting(
                            "SURFEX#COVER#" + decadal_data_type, check_parsing=False
                        )
                        fname = ecoclimap.parse_fnames(filepattern, decade)
                        self.input_list.append(
                            self.set_dirtyp_data_namelist(
                                "NAM_DATA_ISBA",
                                decadal_data_type,
                                fname,
                                vtype=vtt,
                                decade=decade,
                            )
                        )

        ecoclimap_dir = "ecoclimap_dir"
        if self.config.get_setting("SURFEX#COVER#SG"):
            ecoclimap_dir = "ecoclimap_sg_cover_dir"

        possible_direct_data = {
            "ISBA": {
                "YSAND": "sand_dir",
                "YCLAY": "clay_dir",
                "YSOC_TOP": "soc_top_dir",
                "YSOC_SUB": "soc_sub_dir",
            },
            "COVER": {"YCOVER": ecoclimap_dir},
            "ZS": {"YZS": "oro_dir"},
        }
        for namelist_section, ftypes in possible_direct_data.items():
            # for ftype in possible_direct_data[namelist_section]:
            for ftype in ftypes:
                fname = str(
                    self.config.get_setting("SURFEX#" + namelist_section + "#" + ftype)
                )
                self.input_list.append(
                    self.set_direct_data_namelist(
                        "NAM_" + namelist_section, ftype, fname, self.input_path
                    )
                )

        # Set ISBA properties
        if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
            self.input_list.append(
                {"json": {"NAM_ISBA": {"CISBA": "DIF", "NGROUND_LAYER": 14}}}
            )
            if os.path.exists(self.input_path + "/isba_dif.json"):
                self.input_list.append({"file": self.input_path + "/isba_dif.json"})
        elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "3-L":
            self.input_list.append(
                {"json": {"NAM_ISBA": {"CISBA": "3-L", "NGROUND_LAYER": 3}}}
            )
        elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "2-L":
            self.input_list.append(
                {"json": {"NAM_ISBA": {"CISBA": "2-L", "NGROUND_LAYER": 2}}}
            )

        # Set patches
        npatch = self.config.get_setting("SURFEX#ISBA#NPATCH")
        self.input_list.append({"json": {"NAM_ISBA": {"NPATCH": npatch}}})

        # Set MEB
        lmeb = self.config.get_setting("SURFEX#ISBA#MEB")
        self.input_list.append({"json": {"NAM_ISBA": {"LMEB": lmeb}}})
        if lmeb:
            self.input_list.append({"file": self.input_path + "/meb_settings.json"})

        # RSMIN
        if self.config.get_setting("SURFEX#COVER#SG"):
            self.input_list.append({"file": self.input_path + "/rsmin_sg.json"})
            self.input_list.append({"file": self.input_path + "/rsmin_sg_mod.json"})
        else:
            self.input_list.append({"file": self.input_path + "/rsmin.json"})
            self.input_list.append({"file": self.input_path + "/rsmin_mod.json"})

        # CV
        if self.config.get_setting("SURFEX#COVER#SG"):
            self.input_list.append({"file": self.input_path + "/cv_sg.json"})
        else:
            self.input_list.append({"file": self.input_path + "/cv.json"})

        # Treedrag
        if self.config.get_setting("SURFEX#TREEDRAG#TREEDATA_FILE") != "":
            treeheight = self.config.get_setting("SURFEX#TREEDRAG#TREEDATA_FILE")
            self.input_list.append(
                {
                    "json": {
                        "NAM_DATA_ISBA": {
                            "CFNAM_H_TREE(4)": treeheight,
                            "CFTYP_H_TREE(4)": "ASCLLV",
                            "CFNAM_H_TREE(5)": treeheight,
                            "CFTYP_H_TREE(5)": "ASCLLV",
                            "CFNAM_H_TREE(6)": treeheight,
                            "CFTYP_H_TREE(6)": "ASCLLV",
                        }
                    }
                }
            )

        if self.config.get_setting("SURFEX#TOWN#LTOWN_TO_ROCK"):
            if self.config.get_setting("SURFEX#TILES#TOWN") != "NONE":
                logging.warning(
                    "WARNING: TOWN is not NONE and you want LTOWN_TO_ROCK. "
                    "Setting it to NONE!"
                )
            self.input_list.append(
                {"json": {"NAM_PGD_ARRANGE_COVER": {"LTOWN_TO_ROCK": True}}}
            )
            self.input_list.append({"json": {"NAM_PGD_SCHEMES": {"TOWN": "NONE"}}})

        if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
            self.input_list.append(
                {
                    "json": {
                        "NAM_DATA_FLAKE": {
                            "YWATER_DEPTH": "GlobalLakeDepth",
                            "YWATER_DEPTHFILETYPE": "DIRECT",
                            "YWATER_DEPTH_STATUS": "GlobalLakeStatus",
                        }
                    }
                }
            )

        # Sea
        self.input_list.append({"file": self.input_path + "/sea.json"})

    def set_prep_namelist(
        self, prep_file=None, prep_filetype=None, prep_pgdfile=None, prep_pgdfiletype=None
    ):
        """Set prep namelist.

        Args:
            prep_file (_type_, optional): _description_. Defaults to None.
            prep_filetype (_type_, optional): _description_. Defaults to None.
            prep_pgdfile (_type_, optional): _description_. Defaults to None.
            prep_pgdfiletype (_type_, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: Filetype for input to PREP is not set!
            RuntimeError: Filetype for PGD input to PREP is not set!

        """
        if prep_file is not None and prep_filetype is None:
            raise RuntimeError("Filetype for input to PREP is not set!")
        if prep_pgdfile is not None and prep_pgdfiletype is None:
            raise RuntimeError("Filetype for PGD input to PREP is not set!")

        self.input_list.append({"file": self.input_path + "/prep.json"})
        if prep_file is not None:
            if prep_file.endswith(".json"):
                self.input_list.append({"file": prep_file})
            else:
                fname = os.path.basename(prep_file)
                self.input_list.append({"json": {"NAM_PREP_SURF_ATM": {"CFILE": fname}}})
                self.input_list.append(
                    {"json": {"NAM_PREP_SURF_ATM": {"CFILETYPE": prep_filetype}}}
                )
                if prep_pgdfile is not None:
                    fname = os.path.basename(prep_pgdfile)
                    self.input_list.append(
                        {"json": {"NAM_PREP_SURF_ATM": {"CFILEPGD": fname}}}
                    )
                    self.input_list.append(
                        {
                            "json": {
                                "NAM_PREP_SURF_ATM": {"CFILEPGDTYPE": prep_pgdfiletype}
                            }
                        }
                    )
        if self.dtg is not None:
            prep_time = self.dtg
            self.input_list.append(
                {"json": {"NAM_PREP_SURF_ATM": {"NYEAR": int(prep_time.strftime("%Y"))}}}
            )
            self.input_list.append(
                {"json": {"NAM_PREP_SURF_ATM": {"NMONTH": int(prep_time.strftime("%m"))}}}
            )
            self.input_list.append(
                {"json": {"NAM_PREP_SURF_ATM": {"NDAY": int(prep_time.strftime("%d"))}}}
            )
            self.input_list.append(
                {
                    "json": {
                        "NAM_PREP_SURF_ATM": {
                            "XTIME": float(prep_time.strftime("%H")) * 3600.0
                        }
                    }
                }
            )
        else:
            raise RuntimeError("You must provide a DTG for prep")
        if self.config.get_setting("SURFEX#SEA#ICE") == "SICE":
            self.input_list.append(
                {"json": {"NAM_PREP_SEAFLUX": {"CSEAICE_SCHEME": "SICE"}}}
            )
            self.input_list.append({"file": self.input_path + "/prep_sice.json"})

        if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
            lclim_lake = self.config.get_setting("SURFEX#FLAKE#LCLIM")
            self.input_list.append(
                {"json": {"NAM_PREP_FLAKE": {"LCLIM_LAKE": lclim_lake}}}
            )

        # Set extra ISBA-DIF properties (Not needed in prep?)
        if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
            if os.path.exists(self.input_path + "/isba_dif.json"):
                self.input_list.append({"file": self.input_path + "/isba_dif.json"})

        # ISBA CANOPY
        lisba_canopy = self.config.get_setting("SURFEX#ISBA#CANOPY")
        self.input_list.append(
            {"json": {"NAM_PREP_ISBA": {"LISBA_CANOPY": lisba_canopy}}}
        )

        # Snow
        self.input_list.append({"file": self.input_path + "/prep_snow.json"})
        if self.config.get_setting("SURFEX#ISBA#SNOW") == "D95":
            self.input_list.append({"json": {"NAM_PREP_ISBA_SNOW": {"CSNOW": "D95"}}})
        elif self.config.get_setting("SURFEX#ISBA#SNOW") == "3-L":
            self.input_list.append({"json": {"NAM_PREP_ISBA_SNOW": {"CSNOW": "3-L"}}})
        if self.config.get_setting("SURFEX#ISBA#SNOW") == "CRO":
            self.input_list.append({"file": self.input_path + "/snow_crocus.json"})

    def set_offline_namelist(self):
        """Set offline namelist.

        Raises:
            RuntimeError: Mismatch in nnco/cobs_m

        """
        self.input_list.append({"file": self.input_path + "/offline.json"})

        if self.config.get_setting("SURFEX#IO#LSELECT"):
            self.input_list.append({"file": self.input_path + "/selected_output.json"})

        # SEAFLX settings
        if self.config.get_setting("SURFEX#TILES#SEA") == "SEAFLX":
            # Surface perturbations
            pertflux = self.config.get_setting("SURFEX#SEA#PERTFLUX")
            self.input_list.append({"json": {"NAM_SEAFLUXn": {"LPERTFLUX": pertflux}}})

        # ISBA settings
        if self.config.get_setting("SURFEX#TILES#NATURE") == "ISBA":
            pertsurf = self.config.get_setting("SURFEX#ISBA#PERTSURF")
            self.input_list.append({"json": {"NAM_ISBAn": {"LPERTSURF": pertsurf}}})
            xcgmax = self.config.get_setting("SURFEX#ISBA#XCGMAX", abort=False)
            if xcgmax is not None:
                self.input_list.append({"json": {"NAM_ISBAn": {"XCGMAX": xcgmax}}})
            xcsmax = self.config.get_setting("SURFEX#ISBA#XCSMAX", abort=False)
            if xcsmax is not None:
                self.input_list.append({"json": {"NAM_ISBAn": {"XCSMAX": xcsmax}}})

        # SSO
        sso = self.config.get_setting("SURFEX#SSO#SCHEME")
        self.input_list.append({"json": {"NAM_SSON": {"CROUGH": sso}}})
        if sso == "OROTUR":
            self.input_list.append({"json": {"NAM_SSON": {"XSOROT": self.geo.xdx}}})

        # Perturbed offline settings
        self.input_list.append({"json": {"NAM_VAR": {"NIVAR": 0}}})
        self.input_list.append({"json": {"NAM_IO_VARASSIM": {"LPRT": False}}})
        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "EKF":
            cassim_isba = self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA")
            self.input_list.append({"json": {"NAM_ASSIM": {"CASSIM_ISBA": cassim_isba}}})
            nvar = 0
            cvar_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#CVAR_M")
            nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#NNCV")
            xtprt_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XTPRT_M")
            xsigma_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XSIGMA_M")
            for var, cvar_val in enumerate(cvar_m):
                self.input_list.append(
                    {"json": {"NAM_VAR": {"CVAR_M(" + str(var + 1) + ")": cvar_val}}}
                )
                self.input_list.append(
                    {"json": {"NAM_VAR": {"NNCV(" + str(var + 1) + ")": nncv[var]}}}
                )
                self.input_list.append(
                    {"json": {"NAM_VAR": {"XTPRT_M(" + str(var + 1) + ")": xtprt_m[var]}}}
                )
                self.input_list.append(
                    {
                        "json": {
                            "NAM_VAR": {"XSIGMA_M(" + str(var + 1) + ")": xsigma_m[var]}
                        }
                    }
                )
                if nncv[var] == 1:
                    nvar += 1
            self.input_list.append({"json": {"NAM_VAR": {"NVAR": nvar}}})

        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "ENKF":
            self.input_list.append(
                {
                    "json": {
                        "NAM_ASSIM": {
                            "CASSIM_ISBA": self.config.get_setting(
                                "SURFEX#ASSIM#SCHEMES#ISBA"
                            )
                        }
                    }
                }
            )
            nvar = 0
            cvar_m = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#CVAR_M")
            nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#NNCV")
            for var, cvar_val in enumerate(cvar_m):
                self.input_list.append(
                    {"json": {"NAM_VAR": {"CVAR_M(" + str(var + 1) + ")": cvar_val}}}
                )
                self.input_list.append(
                    {"json": {"NAM_VAR": {"NNCV(" + str(var + 1) + ")": nncv[var]}}}
                )
                if nncv[var] == 1:
                    nvar += 1
            self.input_list.append({"json": {"NAM_VAR": {"NVAR": nvar}}})

        # TODO the need for this in forecast must be removed!
        nobstype = 0
        nnco = self.config.get_setting("SURFEX#ASSIM#OBS#NNCO")
        cobs_m = self.config.get_setting("SURFEX#ASSIM#OBS#COBS_M")
        if len(nnco) != len(cobs_m):
            raise RuntimeError("Mismatch in nnco/cobs_m")
        for obs, obs_val in enumerate(nnco):
            self.input_list.append(
                {"json": {"NAM_OBS": {"NNCO(" + str(obs + 1) + ")": obs_val}}}
            )
            self.input_list.append(
                {"json": {"NAM_OBS": {"COBS_M(" + str(obs + 1) + ")": cobs_m[obs]}}}
            )
            if nnco[obs] == 1:
                nobstype += 1
        self.input_list.append({"json": {"NAM_OBS": {"NOBSTYPE": nobstype}}})

        # Climate setting
        if self.config.get_setting("SURFEX#SEA#LVOLATILE_SIC"):
            self.input_list.append(
                {
                    "json": {
                        "NAM_SEAICEn ": {"LVOLATILE_SIC": True, "XSIC_EFOLDING_TIME": 1.0}
                    }
                }
            )

    def set_soda_namelist(self):
        """Set SODA namelist.

        Raises:
            RuntimeError: Mismatch in nnco/cobs_m/xerrobs_m
            RuntimeError: You must provide a DTG when using a list for snow
            RuntimeError: Mismatch in nncv/cvar_m/xsigma_m/xtprt_m

        """
        self.input_list.append({"file": self.input_path + "/soda.json"})

        self.input_list.append({"json": {"NAM_ASSIM": {"LASSIM": True}}})

        lobsheader = self.config.get_setting("SURFEX#ASSIM#OBS#LOBSHEADER")
        self.input_list.append({"json": {"NAM_OBS": {"LOBSHEADER": lobsheader}}})
        lobsnat = self.config.get_setting("SURFEX#ASSIM#OBS#LOBSNAT")
        self.input_list.append({"json": {"NAM_OBS": {"LOBSNAT": lobsnat}}})
        cfile_format_obs = self.config.get_setting("SURFEX#ASSIM#OBS#CFILE_FORMAT_OBS")
        self.input_list.append(
            {"json": {"NAM_OBS": {"CFILE_FORMAT_OBS": cfile_format_obs}}}
        )
        nobstype = 0
        nnco = self.config.get_setting("SURFEX#ASSIM#OBS#NNCO")
        cobs_m = self.config.get_setting("SURFEX#ASSIM#OBS#COBS_M")
        xerrobs_m = self.config.get_setting("SURFEX#ASSIM#OBS#XERROBS_M")
        logging.debug("%s %s %s", nnco, cobs_m, xerrobs_m)
        if len(nnco) != len(cobs_m) or len(nnco) != len(xerrobs_m):
            raise RuntimeError("Mismatch in nnco/cobs_m/xerrobs_m")

        for obs, obs_val in enumerate(nnco):
            self.input_list.append(
                {"json": {"NAM_OBS": {"NNCO(" + str(obs + 1) + ")": obs_val}}}
            )
            self.input_list.append(
                {"json": {"NAM_OBS": {"COBS_M(" + str(obs + 1) + ")": cobs_m[obs]}}}
            )
            self.input_list.append(
                {"json": {"NAM_OBS": {"XERROBS_M(" + str(obs + 1) + ")": xerrobs_m[obs]}}}
            )
            if nnco[obs] == 1:
                nobstype += 1
        self.input_list.append({"json": {"NAM_OBS": {"NOBSTYPE": nobstype}}})
        self.input_list.append(
            {
                "json": {
                    "NAM_OBS": {"LSWE": self.config.get_setting("SURFEX#ASSIM#OBS#LSWE")}
                }
            }
        )

        # LSM
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "CFILE_FORMAT_LSM": self.config.get_setting(
                            "SURFEX#ASSIM#CFILE_FORMAT_LSM"
                        )
                    }
                }
            }
        )

        # Sea
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "CASSIM_SEA": self.config.get_setting("SURFEX#ASSIM#SCHEMES#SEA")
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "CFILE_FORMAT_SST": self.config.get_setting(
                            "SURFEX#ASSIM#SEA#CFILE_FORMAT_SST"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "LREAD_SST_FROM_FILE": self.config.get_setting(
                            "SURFEX#ASSIM#SEA#LREAD_SST_FROM_FILE"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "LEXTRAP_SEA": self.config.get_setting(
                            "SURFEX#ASSIM#SEA#LEXTRAP_SEA"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "LECSST": self.config.get_setting("SURFEX#ASSIM#SEA#LECSST")
                    }
                }
            }
        )

        # Water
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "CASSIM_WATER": self.config.get_setting(
                            "SURFEX#ASSIM#SCHEMES#INLAND_WATER"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "LWATERTG2": self.config.get_setting(
                            "SURFEX#ASSIM#INLAND_WATER#LWATERTG2"
                        )
                    }
                }
            }
        )
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "LEXTRAP_WATER": self.config.get_setting(
                            "SURFEX#ASSIM#INLAND_WATER#LEXTRAP_WATER"
                        )
                    }
                }
            }
        )

        # Nature
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "CASSIM_ISBA": self.config.get_setting(
                            "SURFEX#ASSIM#SCHEMES#ISBA"
                        )
                    }
                }
            }
        )

        # Snow
        self.input_list.append(
            {
                "json": {
                    "NAM_ASSIM": {
                        "LPATCH1": self.config.get_setting("SURFEX#ASSIM#ISBA#LPATCH1")
                    }
                }
            }
        )

        laesnm = False
        snow_cycles = self.config.get_setting("SURFEX#ASSIM#ISBA#UPDATE_SNOW_CYCLES")
        if len(snow_cycles) > 0:
            logging.debug("%s", self.dtg)
            if self.dtg is not None:
                for cycle in snow_cycles:
                    logging.debug(self.dtg.strftime("%H"))
                    if int(self.dtg.strftime("%H")) == int(cycle):
                        logging.debug("true")
                        laesnm = True
            else:
                raise RuntimeError(
                    "You must provide a DTG when using a list for snow  "
                    "assimilation cycles"
                )
        self.input_list.append({"json": {"NAM_ASSIM": {"LAESNM": laesnm}}})

        # Set OI settings
        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "OI":
            ua_physics = self.config.get_setting("FORECAST#PHYSICS", abort=False)
            if ua_physics is None:
                logging.info("Did not find FORECAST#PHYSICS. Assume arome")
                ua_physics = "arome"
            if ua_physics == "arome":
                self.input_list.append({"json": {"NAM_ASSIM": {"LAROME": True}}})
            elif ua_physics == "alaro":
                self.input_list.append({"json": {"NAM_ASSIM": {"LAROME": False}}})

            self.input_list.append(
                {
                    "json": {
                        "NAM_NACVEG": {
                            "XSIGT2MO": self.config.get_setting(
                                "SURFEX#ASSIM#ISBA#OI#XSIGT2MO"
                            )
                        }
                    }
                }
            )
            self.input_list.append(
                {
                    "json": {
                        "NAM_NACVEG": {
                            "XSIGH2MO": self.config.get_setting(
                                "SURFEX#ASSIM#ISBA#OI#XSIGH2MO"
                            )
                        }
                    }
                }
            )
            self.input_list.append({"json": {"NAM_NACVEG": {"XRCLIMCA": 0.0}}})
            self.input_list.append({"json": {"NAM_NACVEG": {"XRCLISST": 0.05}}})
            self.input_list.append({"json": {"NAM_NACVEG": {"NECHGU": self.fcint}}})
            self.input_list.append({"json": {"NAM_NACVEG": {"LOBS2M": True}}})
            self.input_list.append({"json": {"NAM_NACVEG": {"LOBSWG": False}}})
            f_clim = self.config.get_setting("SURFEX#ASSIM#ISBA#OI#CFILE_FORMAT_CLIM")
            self.input_list.append({"json": {"NAM_ASSIM": {"CFILE_FORMAT_CLIM": f_clim}}})
            f_fg = self.config.get_setting("SURFEX#ASSIM#ISBA#OI#CFILE_FORMAT_FG")
            self.input_list.append({"json": {"NAM_ASSIM": {"CFILE_FORMAT_FG": f_fg}}})

        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "EKF":
            nvar = 0
            llincheck = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#LLINCHECK")
            self.input_list.append({"json": {"NAM_ASSIM": {"LLINCHECK": llincheck}}})
            xalpha = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XALPHA")
            self.input_list.append({"json": {"NAM_VAR": {"XALPHA": xalpha}}})
            cvar_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#CVAR_M")
            xsigma_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XSIGMA_M")
            xtprt_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XTPRT_M")
            nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#NNCV")
            if (
                len(nncv) != len(cvar_m)
                or len(nncv) != len(xsigma_m)
                or len(nncv) != len(xtprt_m)
            ):
                raise RuntimeError("Mismatch in nncv/cvar_m/xsigma_m/xtprt_m")
            for var, cvar_val in enumerate(cvar_m):
                self.input_list.append(
                    {"json": {"NAM_VAR": {"CVAR_M(" + str(var + 1) + ")": cvar_val}}}
                )
                self.input_list.append(
                    {
                        "json": {
                            "NAM_VAR": {"XSIGMA_M(" + str(var + 1) + ")": xsigma_m[var]}
                        }
                    }
                )
                self.input_list.append(
                    {"json": {"NAM_VAR": {"XTPRT_M(" + str(var + 1) + ")": xtprt_m[var]}}}
                )
                self.input_list.append(
                    {"json": {"NAM_VAR": {"NNCV(" + str(var + 1) + ")": nncv[var]}}}
                )
                if nncv[var] == 1:
                    nvar += 1
            self.input_list.append({"json": {"NAM_VAR": {"NIVAR": 0}}})
            self.input_list.append({"json": {"NAM_VAR": {"NVAR": nvar}}})
            xscale_q = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XSCALE_Q")
            self.input_list.append({"json": {"NAM_VAR": {"XSCALE_Q": xscale_q}}})
            self.input_list.append(
                {
                    "json": {
                        "NAM_IO_VARASSIM": {
                            "LPRT": False,
                            "LBEV": self.config.get_setting(
                                "SURFEX#ASSIM#ISBA#EKF#EVOLVE_B"
                            ),
                            "LBFIXED": not self.config.get_setting(
                                "SURFEX#ASSIM#ISBA#EKF#EVOLVE_B"
                            ),
                        }
                    }
                }
            )

        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "ENKF":
            nvar = 0
            cvar_m = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#CVAR_M")
            nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#NNCV")
            if len(nncv) != len(cvar_m):
                raise RuntimeError("Mismatch in nncv/cvar_m")
            for var, cvar_val in enumerate(cvar_m):
                self.input_list.append(
                    {"json": {"NAM_VAR": {"CVAR_M(" + str(var + 1) + ")": cvar_val}}}
                )
                self.input_list.append(
                    {"json": {"NAM_VAR": {"NNCV(" + str(var + 1) + ")": nncv[var]}}}
                )

                if nncv[var] == 1:
                    nvar += 1
            self.input_list.append({"json": {"NAM_VAR": {"NIVAR": 0}}})
            self.input_list.append({"json": {"NAM_VAR": {"NVAR": nvar}}})

        # Town
        town_setting = self.config.get_setting("SURFEX#ASSIM#SCHEMES#TEB")
        self.input_list.append({"json": {"NAM_ASSIM": {"CASSIM_TEB": town_setting}}})

    def epilog(self):
        """Epilog."""
        # Always set these
        if self.config.get_setting("SURFEX#SEA#ICE") == "SICE":
            self.input_list.append({"file": self.input_path + "/sice.json"})

        self.input_list.append({"file": self.input_path + "/treedrag.json"})
        lfaketrees = self.config.get_setting("SURFEX#TREEDRAG#FAKETREES", abort=False)
        if lfaketrees is not None:
            self.input_list.append({"json": {"NAM_TREEDRAG": {"LFAKETREE": lfaketrees}}})

        if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
            self.input_list.append({"file": self.input_path + "/flake.json"})

    def override(self):
        """Overide."""
        # Override posssibility
        if os.path.exists(self.input_path + "/override.json"):
            logging.warning(
                "WARNING: Override settings with content from %s/override.json",
                self.input_path,
            )
            self.input_list.append({"file": self.input_path + "/override.json"})

    @staticmethod
    def set_direct_data_namelist(lnamelist_section, ldtype, ldname, linput_path):
        """Set direct data namelist.

        Args:
            lnamelist_section (_type_): _description_
            ldtype (_type_): _description_
            ldname (_type_): _description_
            linput_path (_type_): _description_

        Returns:
            _type_: _description_

        """
        if ldname.endswith(".dir"):
            basename = os.path.splitext(os.path.basename(ldname))[0]
            filetype_name = ldtype
            if ldtype == "YSOC_TOP" or ldtype == "YSOC_SUB":
                filetype_name = "YSOC"
            return {
                "json": json.loads(
                    '{"'
                    + lnamelist_section
                    + '": { "'
                    + ldtype
                    + '": "'
                    + basename
                    + '", '
                    + '"'
                    + filetype_name
                    + 'FILETYPE": "DIRECT"}}'
                )
            }
        if ldname.endswith(".json"):
            return {"file": linput_path + "/" + ldname}

    @staticmethod
    def set_dirtyp_data_namelist(
        lnamelist_section, ldtype, ldname, vtype=None, decade=None
    ):
        """Set dirtyp data namelist.

        Args:
            lnamelist_section (_type_): _description_
            ldtype (_type_): _description_
            ldname (_type_): _description_
            vtype (_type_, optional): _description_. Defaults to None.
            decade (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_

        """
        basename = os.path.splitext(os.path.basename(ldname))[0]
        filetype_name = ldtype.upper()
        if vtype is not None or decade is not None:
            filetype_name = filetype_name + "("
        if vtype is not None:
            filetype_name = filetype_name + str(vtype)
        if vtype is not None and decade is not None:
            filetype_name = filetype_name + ","
        if decade is not None:
            filetype_name = filetype_name + str(decade)
        if vtype is not None or decade is not None:
            filetype_name = filetype_name + ")"
        return {
            "json": json.loads(
                '{"'
                + lnamelist_section
                + '": { "CFNAM_'
                + filetype_name
                + '": "'
                + basename
                + '", "CFTYP_'
                + filetype_name
                + '": "DIRTYPE"}}'
            )
        }

    @staticmethod
    def capitalize_namelist_dict(dict_in):
        """Capitalize namelist.

        Args:
            dict_in (dict): _description_

        Returns:
            _type_: _description_

        """
        new_dict = {}
        for key in dict_in:
            upper_case2 = {}
            for key2 in dict_in[key]:
                upper_case2.update({key2.upper(): dict_in[key][key2]})
            new_dict.update({key.upper(): upper_case2})
        return new_dict

    @staticmethod
    def lower_case_namelist_dict(dict_in):
        """Lower case namelist.

        Args:
            dict_in (dict): Namelist dictionary to lower case

        Returns:
            dict: Namelist in lower case

        """
        new_dict = {}
        for key in dict_in:
            lower_case_dict = {}
            for key2 in dict_in[key]:
                lower_case_dict.update({key2.lower(): dict_in[key][key2]})
            new_dict.update({key.lower(): lower_case_dict})
        return new_dict

    @staticmethod
    def merge_namelist_dicts(old_dict, new_dict):
        """Merge namelist dicts.

        Args:
            old_dict (dict): Old dictionary
            new_dict (dict): New dictionary

        Returns:
            dict: Merged dictionary

        """
        old_dict = Namelist.capitalize_namelist_dict(old_dict)
        new_dict = Namelist.capitalize_namelist_dict(new_dict)
        merged_dict = old_dict

        for new_key, new_val in new_dict.items():
            # Namelist block already exists
            if new_key in merged_dict:
                settings = merged_dict[new_key]
                for new_key2 in new_val:
                    settings.update({new_key2: new_dict[new_key][new_key2]})

                merged_dict.update({new_key: settings})
            # New namelist block
            else:
                merged_dict.update({new_key: new_val})

        return merged_dict

    @staticmethod
    def ascii2nml(input_data):
        """Convert dict to a namelist object.

        Args:
            input_data (dict): Namelist settings

        Returns:
            f90nml.Namelist: Namelist object.
        """
        output_data = f90nml.Namelist(input_data)
        return output_data

    @staticmethod
    def ascii_file2nml(input_fname, input_fmt="json"):
        """Convert a file wih namelist settings to a Namelist object.

        Args:
            input_fname (str): Filname
            input_fmt (str, optional): File format. Defaults to "json".

        Returns:
            f90nml.Namelist: Namelist object.

        """
        if input_fmt == "json":
            with open(input_fname, mode="r", encoding="utf-8") as input_file:
                output_data = json.load(input_file)
        elif input_fmt == "yaml":
            with open(input_fname, mode="r", encoding="utf-8") as input_file:
                output_data = yaml.safe_load(input_file)
        output_data = f90nml.Namelist(output_data)
        return output_data

    @staticmethod
    def nml2ascii(input_data, output_file, output_fmt="json", indent=2):
        """Dump a namelist object as a dict in a json or yaml file.

        Args:
            input_data (f90nml.Namelist): Namelist object.
            output_file (str): Filename
            output_fmt (str, optional): File format. Defaults to "json".
            indent (int, optional): Indentation. Defaults to 2.

        """
        if output_fmt == "json":
            input_data = input_data.todict(complex_tuple=True)
            json.dump(
                input_data,
                open(output_file, "w", encoding="utf-8"),
                indent=indent,
                separators=(",", ": "),
            )
        elif output_fmt == "yaml":
            input_data = input_data.todict(complex_tuple=True)
            yaml.dump(input_data, output_file, default_flow_style=False)

    @staticmethod
    def merge_json_namelist_file(old_dict, my_file):
        """Merge json files with namelists.

        Args:
            old_dict (dict): Exististing settings
            my_file (str): Filename with new settings

        Raises:
            FileNotFoundError: Namelist input not found

        Returns:
            dict: Merged settings.

        """
        logging.debug(my_file)
        if os.path.exists(my_file):
            with open(my_file, "r", encoding="utf-8") as file_handler:
                new_dict = json.load(file_handler)
        else:
            raise FileNotFoundError(my_file)

        return Namelist.merge_namelist_dicts(old_dict, new_dict)

    def get_namelist(self):
        """Get namelist."""
        logging.debug("Constructing namelist:")
        merged_json_settings = {}
        for inp in self.input_list:
            if "file" in inp:
                json_file = str(inp["file"])
                if not os.path.exists(json_file):
                    raise FileNotFoundError(
                        "Needed namelist input does not exist: " + json_file
                    )
                else:
                    merged_json_settings = self.merge_json_namelist_file(
                        merged_json_settings, json_file
                    )
            elif "json" in inp:
                merged_json_settings = self.merge_namelist_dicts(
                    merged_json_settings, inp["json"]
                )
            else:
                logging.error("Can not handle input type %s", str(inp))
                raise Exception

        return self.ascii2nml(merged_json_settings)


class Namelist(object):
    """Base namelist."""

    def __init__(
        self,
        program,
        config,
        input_path,
        forc_zs=False,
        prep_file=None,
        prep_filetype=None,
        prep_pgdfile=None,
        prep_pgdfiletype=None,
        dtg=None,
        fcint=3,
        geo=None,
    ):
        """Construct a base namelists class to be implemented by namelist implementations.

        Args:
            program (str): Which surfex binary you want to run ["pgd", "prep", "offline", "soda"]
            config (surfex.Configuration): A SURFEX configuration object
            input_path (str): A path to search for json namelist blocks
            forc_zs (bool): Set the surfex orography heigth to the same as in the forcing files
            prep_file (str): Input file for prep
            prep_filetype (str): The format of prep_file
            prep_pgdfile (str): Input PGD file for prep in case it comes from SURFEX
            prep_pgdfiletype (str): The format of the prep_pgdfile
            dtg (datetime.datetime): The date/time you want to run
            fcint (int): The intervall between the cycles. Used for first guesses.
            geo (surfex.Geo): Surfex geometry. The domain you want to run on

        Raises:
            RuntimeError: Input
            RuntimeError: Merged dictionary contains a @ in value
            NotImplementedError: Mode is not implemented

        """
        self.config = config
        self.input_path = input_path
        self.forc_zs = forc_zs
        if dtg is not None:
            if isinstance(dtg, str):
                dtg = as_datetime(dtg)
        self.dtg = dtg
        check_parsing = True
        if self.dtg is None:
            check_parsing = False
        self.fcint = fcint
        self.geo = geo

        # The time stamp of next cycle file
        forecast_length = self.fcint
        if self.dtg is not None:
            self.end_of_forecast = self.dtg + as_timedelta(seconds=forecast_length * 3600)
        else:
            self.end_of_forecast = None

        logging.info("Creating JSON namelist input for program: %s", program)

        merged_dict = {}
        merged_dict = self.prolog(merged_dict, check_parsing=check_parsing)
        # Program specific settings
        if program == "pgd":
            merged_dict = self.set_pgd_namelist(merged_dict)
        elif program == "prep":
            if prep_file is None:
                raise RuntimeError(
                    "Prep need an input file either as a json namelist or a surfex "
                    "supported format"
                )
            merged_dict = self.set_prep_namelist(
                merged_dict,
                prep_file=prep_file,
                prep_filetype=prep_filetype,
                prep_pgdfile=prep_pgdfile,
                prep_pgdfiletype=prep_pgdfiletype,
            )
        elif program == "offline" or program == "perturbed":
            merged_dict = self.set_offline_namelist(merged_dict)
        elif program == "soda":
            merged_dict = self.set_soda_namelist(merged_dict)
        else:
            raise NotImplementedError(program)
        merged_dict = self.epilog(merged_dict)
        merged_dict = self.override(merged_dict)

        # Check parsing
        non_parsed = {}
        for block in merged_dict:
            keys = []
            for key, value in merged_dict[block].items():
                if isinstance(value, str) and value[0] == "@" and value[-1] == "@":
                    logging.info(
                        "Delete non-substituted placeholder %s for key %s", value, key
                    )
                    keys.append(key)
            if len(keys) > 0:
                non_parsed.update({block: keys})

        for block, keys in non_parsed.items():
            for key in keys:
                merged_dict = self.delete(merged_dict, block, key)

        for block in merged_dict:
            for key, value in merged_dict[block].items():
                if isinstance(value, str) and "@" in value[0] and "@" in value[-1]:
                    logging.error(
                        "Merged dictionary contains a @ in value %s for key %s",
                        value,
                        key,
                    )
                    raise RuntimeError("Merged dictionary contains a @ in value")

        self.namelist_dict = merged_dict

    def prolog(self, merged_dict, check_parsing=True):
        """Prolog.

        Args:
            merged_dict(dict): Merged settings.
            check_parsing (bool, optional): Check if parsing is ok. Defaults to True.

        Returns:
            merged_dict(dict): Merged settings

        """
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/prolog.json"
        )
        # IO
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/io.json"
        )
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/constants.json"
        )

        # IO manipulation
        merged_dict = self.sub(
            merged_dict,
            "NAM_IO_OFFLINE",
            "CSURF_FILETYPE",
            self.config.get_setting("SURFEX#IO#CSURF_FILETYPE"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_IO_OFFLINE",
            "CTIMESERIES_FILETYPE",
            self.config.get_setting("SURFEX#IO#CTIMESERIES_FILETYPE"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_IO_OFFLINE",
            "CFORCING_FILETYPE",
            self.config.get_setting("SURFEX#IO#CFORCING_FILETYPE"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_IO_OFFLINE",
            "CPGDFILE",
            self.config.get_setting("SURFEX#IO#CPGDFILE"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_IO_OFFLINE",
            "CPREPFILE",
            self.config.get_setting("SURFEX#IO#CPREPFILE"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_IO_OFFLINE",
            "CSURFFILE",
            self.config.get_setting(
                "SURFEX#IO#CSURFFILE",
                validtime=self.end_of_forecast,
                basedtg=self.dtg,
                check_parsing=check_parsing,
            ),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_IO_OFFLINE",
            "XTSTEP_SURF",
            self.config.get_setting("SURFEX#IO#XTSTEP"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_IO_OFFLINE",
            "XTSTEP_OUTPUT",
            self.config.get_setting("SURFEX#IO#XTSTEP_OUTPUT"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_WRITE_SURF_ATM",
            "LSPLIT_PATCH",
            self.config.get_setting("SURFEX#IO#LSPLIT_PATCH"),
        )
        merged_dict = self.sub(
            merged_dict, "NAM_IO_OFFLINE", "LSET_FORC_ZS", self.forc_zs
        )

        # Constants and parameters
        merged_dict = self.sub(
            merged_dict,
            "NAM_SURF_ATM",
            "XRIMAX",
            self.config.get_setting("SURFEX#PARAMETERS#XRIMAX"),
        )
        return merged_dict

    def set_pgd_namelist(self, merged_dict):
        """Set pgd namelist."""
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/pgd.json"
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_PGD_SCHEMES",
            "CSEA",
            self.config.get_setting("SURFEX#TILES#SEA"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_PGD_SCHEMES",
            "CWATER",
            self.config.get_setting("SURFEX#TILES#INLAND_WATER"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_PGD_SCHEMES",
            "CNATURE",
            self.config.get_setting("SURFEX#TILES#NATURE"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_PGD_SCHEMES",
            "CTOWN",
            self.config.get_setting("SURFEX#TILES#TOWN"),
        )
        if self.config.get_setting("SURFEX#TOWN#LTOWN_TO_ROCK"):

            if self.config.get_setting("SURFEX#TILES#TOWN") != "NONE":
                logging.warning(
                    "WARNING: TOWN is not NONE and you want LTOWN_TO_ROCK. "
                    "Setting it to NONE!"
                )
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_arrange_cover.json"
            )
            merged_dict = self.sub(
                merged_dict, "NAM_PGD_ARRANGE_COVER", "LTOWN_TO_ROCK", True
            )
            merged_dict = self.sub(merged_dict, "NAM_PGD_SCHEMES", "CTOWN", "NONE")

        # Cover
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/pgd_cover.json"
        )

        # COVER
        eco_sg = self.config.get_setting("SURFEX#COVER#SG")
        merged_dict = self.sub(merged_dict, "NAM_FRAC", "LECOSG", eco_sg)
        # Ecoclimap SG
        if eco_sg:
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_eco_sg.json"
            )
            ecoclimap = EcoclimapSG(self.config)
        else:
            ecoclimap = Ecoclimap(self.config)

        fname = str(self.config.get_setting("SURFEX#COVER#YCOVER"))
        fname, filetype = self.get_filetype_from_suffix(fname)
        merged_dict = self.sub(merged_dict, "NAM_COVER", "YCOVER", fname)
        if filetype is not None:
            merged_dict = self.sub(merged_dict, "NAM_COVER", "YCOVERFILETYPE", filetype)

        # ZS
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/pgd_zs.json"
        )
        fname = str(self.config.get_setting("SURFEX#ZS#YZS"))
        fname, filetype = self.get_filetype_from_suffix(fname)
        merged_dict = self.sub(merged_dict, "NAM_ZS", "YZS", fname)
        if filetype is not None:
            merged_dict = self.sub(merged_dict, "NAM_ZS", "YZSFILETYPE", filetype)

        # Sea
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/pgd_sea.json"
        )

        # Inland water
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/pgd_inland_water.json"
        )

        if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_flake.json"
            )

        # PGD ISBA
        if self.config.get_setting("SURFEX#TILES#NATURE") == "ISBA":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_isba.json"
            )

        if eco_sg:
            merged_dict = self.sub(
                merged_dict, "NAM_DATA_ISBA", "NTIME", ecoclimap.decades
            )
            fname = self.config.get_setting("SURFEX#COVER#H_TREE")
            if fname != "" and fname is not None:
                if fname.endswith(".dir"):
                    fname = fname.replace(".dir", "")
                for vtt in range(1, ecoclimap.veg_types + 1):
                    merged_dict = self.sub(
                        merged_dict,
                        "NAM_DATA_ISBA",
                        "CFNAM_H_TREE(@VEGTYPE@)",
                        fname,
                        vtype=vtt,
                    )
                    merged_dict = self.sub(
                        merged_dict,
                        "NAM_DATA_ISBA",
                        "CFTYP_H_TREE(@VEGTYPE@)",
                        "DIRTYPE",
                        vtype=vtt,
                    )

                key = "CFNAM_H_TREE(@VEGTYPE@)"
                merged_dict = self.delete(merged_dict, "NAM_DATA_ISBA", key)
                key = "CFTYP_H_TREE(@VEGTYPE@)"
                merged_dict = self.delete(merged_dict, "NAM_DATA_ISBA", key)

            decadal_data_types = [
                "ALBNIR_SOIL",
                "ALBNIR_VEG",
                "ALBVIS_SOIL",
                "ALBVIS_VEG",
                "LAI",
            ]
            for decadal_data_type in decadal_data_types:
                for vtt in range(1, ecoclimap.veg_types + 1):
                    for decade in range(1, ecoclimap.decades + 1):
                        filepattern = self.config.get_setting(
                            "SURFEX#COVER#" + decadal_data_type, check_parsing=False
                        )
                        fname = ecoclimap.parse_fnames(filepattern, decade)
                        if fname.endswith(".dir"):
                            fname = fname.replace(".dir", "")
                        key = f"CFNAM_{decadal_data_type}(@VEGTYPE@,@DECADE@)"
                        merged_dict = self.sub(
                            merged_dict,
                            "NAM_DATA_ISBA",
                            key,
                            fname,
                            vtype=vtt,
                            decade=decade,
                        )
                        key = f"CFTYP_{decadal_data_type}(@VEGTYPE@,@DECADE@)"
                        merged_dict = self.sub(
                            merged_dict,
                            "NAM_DATA_ISBA",
                            key,
                            "DIRTYPE",
                            vtype=vtt,
                            decade=decade,
                        )

                # Delete generic placehoder
                key = f"CFNAM_{decadal_data_type}(@VEGTYPE@,@DECADE@)"
                merged_dict = self.delete(merged_dict, "NAM_DATA_ISBA", key)
                key = f"CFTYP_{decadal_data_type}(@VEGTYPE@,@DECADE@)"
                merged_dict = self.delete(merged_dict, "NAM_DATA_ISBA", key)

        isba_input_data = ["YSAND", "YCLAY", "YSOC_TOP", "YSOC_SUB"]
        for ftype in isba_input_data:
            fname = str(self.config.get_setting("SURFEX#ISBA#" + ftype))
            fname, filetype = self.get_filetype_from_suffix(fname)
            merged_dict = self.sub(merged_dict, "NAM_ISBA", ftype, fname)
            key = ftype + "FILETYPE"
            if ftype in ["YSOC_TOP", "YSOC_SUB"]:
                if ftype in ["YSOC_TOP"]:
                    key = "YSOCFILETYPE"
                    merged_dict = self.sub(merged_dict, "NAM_ISBA", key, filetype)
            else:
                merged_dict = self.sub(merged_dict, "NAM_ISBA", key, filetype)

        # Set ISBA properties
        if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_isba_dif.json"
            )
            merged_dict = self.sub(merged_dict, "NAM_ISBA", "CISBA", "DIF")
            merged_dict = self.sub(merged_dict, "NAM_ISBA", "NGROUND_LAYER", 14)
        elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "3-L":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_isba_3-L.json"
            )
            merged_dict = self.sub(merged_dict, "NAM_ISBA", "CISBA", "3-L")
            merged_dict = self.sub(merged_dict, "NAM_ISBA", "NGROUND_LAYER", 3)
        elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "2-L":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_isba_2-L.json"
            )
            merged_dict = self.sub(merged_dict, "NAM_ISBA", "CISBA", "2-L")
            merged_dict = self.sub(merged_dict, "NAM_ISBA", "NGROUND_LAYER", 2)

        # Set patches
        npatch = self.config.get_setting("SURFEX#ISBA#NPATCH")
        merged_dict = self.sub(merged_dict, "NAM_ISBA", "NPATCH", npatch)

        # Set MEB
        lmeb = self.config.get_setting("SURFEX#ISBA#MEB")
        merged_dict = self.sub(merged_dict, "NAM_ISBA", "LMEB", lmeb)
        if lmeb:
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/meb_settings.json"
            )

        # RSMIN
        if self.config.get_setting("SURFEX#COVER#SG"):
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_rsmin_sg.json"
            )
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_rsmin_sg_mod.json"
            )
        else:
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_rsmin.json"
            )
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_rsmin_mod.json"
            )

        # CV
        if self.config.get_setting("SURFEX#COVER#SG"):
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_cv_sg.json"
            )
        else:
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/pgd_cv.json"
            )

        # Treedrag (from ASCLLV files)
        if self.config.get_setting("SURFEX#TREEDRAG#TREEDATA_FILE") != "":
            treeheight = self.config.get_setting("SURFEX#TREEDRAG#TREEDATA_FILE")
            for vtt in [4, 5, 6]:
                merged_dict = self.sub(
                    merged_dict,
                    "NAM_DATA_ISBA",
                    "CFNAM_H_TREE(@VEGTYPE@)",
                    treeheight,
                    vtype=vtt,
                )
                merged_dict = self.sub(
                    merged_dict,
                    "NAM_DATA_ISBA",
                    "CFNAM_H_TREE(@VEGTYPE@)",
                    "ASCLLV",
                    vtype=vtt,
                )

            key = "CFNAM_H_TREE(@VEGTYPE@)"
            merged_dict = self.delete(merged_dict, "NAM_DATA_ISBA", key)
            key = "CFTYP_H_TREE(@VEGTYPE@)"
            merged_dict = self.delete(merged_dict, "NAM_DATA_ISBA", key)

        return merged_dict

    def set_prep_namelist(
        self,
        merged_dict,
        prep_file=None,
        prep_filetype=None,
        prep_pgdfile=None,
        prep_pgdfiletype=None,
    ):
        """Set prep namelist.

        Args:
            merged_dict(dict): Merged settings.
            prep_file (_type_, optional): _description_. Defaults to None.
            prep_filetype (_type_, optional): _description_. Defaults to None.
            prep_pgdfile (_type_, optional): _description_. Defaults to None.
            prep_pgdfiletype (_type_, optional): _description_. Defaults to None.

        Raises:
            FileNotFoundError: Main prep namelist input
            RuntimeError: You must provide a DTG for prep
            RuntimeError: Filetype for input to PREP is not set!
            RuntimeError: Filetype for PGD input to PREP is not set

        Returns:
            merged_dict(dict): Merged settings.

        """
        if prep_file is not None and prep_filetype is None:
            raise RuntimeError("Filetype for input to PREP is not set!")
        if prep_pgdfile is not None and prep_pgdfiletype is None:
            raise RuntimeError("Filetype for PGD input to PREP is not set!")

        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/prep.json"
        )
        if prep_file is not None:
            if prep_file.endswith(".json"):
                if not os.path.exists(prep_file):
                    raise FileNotFoundError(f"{prep_file} does not exist!")
                merged_dict = self.merge_json_namelist_file(merged_dict, prep_file)
            else:
                fname = os.path.basename(prep_file)
                merged_dict = self.sub(merged_dict, "NAM_PREP_SURF_ATM", "CFILE", fname)
                merged_dict = self.sub(
                    merged_dict, "NAM_PREP_SURF_ATM", "CFILETYPE", prep_filetype
                )
                if prep_pgdfile is not None:
                    fname = os.path.basename(prep_pgdfile)
                    merged_dict = self.sub(
                        merged_dict, "NAM_PREP_SURF_ATM", "CFILEPGD", fname
                    )
                    merged_dict = self.sub(
                        merged_dict, "NAM_PREP_SURF_ATM", "CFILEPGDTYPE", prep_pgdfiletype
                    )
                else:
                    merged_dict = self.delete(
                        merged_dict, "NAM_PREP_SURF_ATM", "CFILEPGD"
                    )
                    merged_dict = self.delete(
                        merged_dict, "NAM_PREP_SURF_ATM", "CFILEPGDTYPE"
                    )

        if self.dtg is not None:
            prep_time = self.dtg
            merged_dict = self.sub(
                merged_dict, "NAM_PREP_SURF_ATM", "NYEAR", int(prep_time.strftime("%Y"))
            )
            merged_dict = self.sub(
                merged_dict, "NAM_PREP_SURF_ATM", "NMONTH", int(prep_time.strftime("%m"))
            )
            merged_dict = self.sub(
                merged_dict, "NAM_PREP_SURF_ATM", "NDAY", int(prep_time.strftime("%d"))
            )
            merged_dict = self.sub(
                merged_dict,
                "NAM_PREP_SURF_ATM",
                "XTIME",
                float(prep_time.strftime("%H")) * 3600.0,
            )
        else:
            raise RuntimeError("You must provide a DTG for prep")

        if self.config.get_setting("SURFEX#TILES#SEA") == "SEAFLX":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/prep_seaflux.json"
            )
            sea_ice = self.config.get_setting("SURFEX#SEA#ICE")
            merged_dict = self.sub(
                merged_dict, "NAM_PREP_SEAFLUX", "CSEAICE_SCHEME", sea_ice
            )
            if sea_ice == "SICE":
                merged_dict = self.merge_json_namelist_file(
                    merged_dict, self.input_path + "/prep_sice.json"
                )

        if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/prep_flake.json"
            )
            lclim_lake = self.config.get_setting("SURFEX#FLAKE#LCLIM")
            merged_dict = self.sub(
                merged_dict, "NAM_PREP_FLAKE", "LCLIM_LAKE", lclim_lake
            )

        if self.config.get_setting("SURFEX#TILES#NATURE") == "ISBA":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/prep_isba.json"
            )
            # Set extra ISBA-DIF properties (Not needed in prep?)
            if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
                merged_dict = self.merge_json_namelist_file(
                    merged_dict, self.input_path + "/prep_isba_dif.json"
                )

            # ISBA CANOPY
            lisba_canopy = self.config.get_setting("SURFEX#ISBA#CANOPY")
            merged_dict = self.sub(
                merged_dict, "NAM_PREP_ISBA", "LISBA_CANOPY", lisba_canopy
            )

            # Snow
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/prep_isba_snow.json"
            )
            snow_scheme = self.config.get_setting("SURFEX#ISBA#SNOW")
            merged_dict = self.sub(
                merged_dict, "NAM_PREP_ISBA_SNOW", "CSNOW", snow_scheme
            )
            if self.config.get_setting("SURFEX#ISBA#SNOW") == "CRO":
                merged_dict = self.merge_json_namelist_file(
                    merged_dict, self.input_path + "/prep_isba_snow_crocus.json"
                )

        return merged_dict

    def set_offline_namelist(self, merged_dict):
        """Set offline namelist."""
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/offline.json"
        )

        if self.config.get_setting("SURFEX#IO#LSELECT"):
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/selected_output.json"
            )

        # SEAFLX settings
        if self.config.get_setting("SURFEX#TILES#SEA") == "SEAFLX":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_seaflux.json"
            )
            # Surface perturbations
            pertflux = self.config.get_setting("SURFEX#SEA#PERTFLUX")
            merged_dict = self.sub(merged_dict, "NAM_SEAFLUXn", "LPERTFLUX", pertflux)

        if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "WATFLX":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_watflux.json"
            )

        if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_flake.json"
            )

        # ISBA settings
        if self.config.get_setting("SURFEX#TILES#NATURE") == "ISBA":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_isba.json"
            )
            pertsurf = self.config.get_setting("SURFEX#ISBA#PERTSURF")
            merged_dict = self.sub(merged_dict, "NAM_ISBAn", "LPERTSURF", pertsurf)
            xcgmax = self.config.get_setting("SURFEX#ISBA#XCGMAX", abort=False)
            if xcgmax is not None:
                merged_dict = self.sub(merged_dict, "NAM_ISBAn", "XCGMAX", xcgmax)
            xcsmax = self.config.get_setting("SURFEX#ISBA#XCSMAX", abort=False)
            if xcsmax is not None:
                merged_dict = self.sub(merged_dict, "NAM_ISBAn", "XCSMAX", xcsmax)

        # SSO
        sso = self.config.get_setting("SURFEX#SSO#SCHEME")
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/offline_sso.json"
        )
        merged_dict = self.sub(merged_dict, "NAM_SSON", "CROUGH", sso)
        if sso == "OROTUR":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_sso_orotur.json"
            )
            merged_dict = self.sub(merged_dict, "NAM_ISBAn", "XSOROT", self.geo.xdx)

        if self.config.get_setting("SURFEX#TILES#TOWN") == "TEB":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_teb.json"
            )

        # Perturbed offline settings (are overridden when running binary)
        # Make sure variables are existing and consistent
        merged_dict = self.prepare_offline_perturbation(merged_dict)

        # TODO the need for this in forecast must be removed!
        merged_dict = self.set_obs(merged_dict)

        # Climate setting
        if self.config.get_setting("SURFEX#SEA#LVOLATILE_SIC"):
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_seaice.json"
            )
            merged_dict = self.sub(merged_dict, "NAM_SEAICEn", "LVOLATILE_SIC", True)
            merged_dict = self.sub(merged_dict, "NAM_SEAICEn", "XSIC_EFOLDING_TIME", 1.0)

        return merged_dict

    def prepare_offline_perturbation(self, merged_dict):
        """Prepare for offline pertubations.

        Args:
            merged_dict (_type_): _description_

        Returns:
            _type_: _description_

        """
        if "NAM_VAR" not in merged_dict:
            merged_dict.update({"NAM_VAR": {}})
        if "NAM_IO_VARASSIM" not in merged_dict:
            merged_dict.update({"NAM_IO_VARASSIM": {}})
        merged_dict["NAM_VAR"].update({"NIVAR": 0})
        merged_dict["NAM_IO_VARASSIM"].update({"LPRT": False})
        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "EKF":

            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_assim_pert.json"
            )

            merged_dict = self.sub(merged_dict, "NAM_ASSIM", "CASSIM_ISBA", "EKF")
            nvar = 0
            cvar_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#CVAR_M")
            nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#NNCV")
            xtprt_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XTPRT_M")
            xsigma_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XSIGMA_M")
            for var, cvar_val in enumerate(cvar_m):
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "CVAR_M(@VAR@)", cvar_val, var=var + 1
                )
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "NNCV(@VAR@)", nncv[var], var=var + 1
                )
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "XTPRT_M(@VAR@)", xtprt_m[var], var=var + 1
                )
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "XSIGMA_M(@VAR@)", xsigma_m[var], var=var + 1
                )
                if nncv[var] == 1:
                    nvar += 1
            merged_dict = self.delete(merged_dict, "NAM_VAR", "XSIGMA_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "XTPRT_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "CVAR_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "NNCV(@VAR@)")
            merged_dict = self.sub(merged_dict, "NAM_VAR", "NVAR", nvar)

        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "ENKF":

            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/offline_assim_pert.json"
            )
            merged_dict = self.sub(merged_dict, "NAM_ASSIM", "CASSIM_ISBA", "ENKF")

            nvar = 0
            cvar_m = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#CVAR_M")
            nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#NNCV")
            for var, cvar_val in enumerate(cvar_m):
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "CVAR_M(@VAR@)", cvar_val, var=var + 1
                )
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "NNCV(@VAR@)", nncv[var], var=var + 1
                )
                if nncv[var] == 1:
                    nvar += 1
            merged_dict = self.delete(merged_dict, "NAM_VAR", "XSIGMA_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "XTPRT_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "CVAR_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "NNCV(@VAR@)")
            merged_dict = self.sub(merged_dict, "NAM_VAR", "NVAR", nvar)
        return merged_dict

    def set_obs(self, merged_dict):
        """Set obs.

        Args:
            merged_dict (dict): Merged settings

        Raises:
            RuntimeError: Mismatch in nnco/cobs_m/xerrobs_m

        Returns:
            _type_: _description_

        """
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/soda_obs.json"
        )
        lobsheader = self.config.get_setting("SURFEX#ASSIM#OBS#LOBSHEADER")
        merged_dict = self.sub(merged_dict, "NAM_OBS", "LOBSHEADER", lobsheader)
        lobsnat = self.config.get_setting("SURFEX#ASSIM#OBS#LOBSNAT")
        merged_dict = self.sub(merged_dict, "NAM_OBS", "LOBSNAT", lobsnat)
        cfile_format_obs = self.config.get_setting("SURFEX#ASSIM#OBS#CFILE_FORMAT_OBS")
        merged_dict = self.sub(
            merged_dict, "NAM_OBS", "CFILE_FORMAT_OBS", cfile_format_obs
        )
        nobstype = 0
        nnco = self.config.get_setting("SURFEX#ASSIM#OBS#NNCO")
        cobs_m = self.config.get_setting("SURFEX#ASSIM#OBS#COBS_M")
        xerrobs_m = self.config.get_setting("SURFEX#ASSIM#OBS#XERROBS_M")
        logging.debug("%s %s %s", nnco, cobs_m, xerrobs_m)
        if len(nnco) != len(cobs_m) or len(nnco) != len(xerrobs_m):
            raise RuntimeError("Mismatch in nnco/cobs_m/xerrobs_m")

        for obs, obs_val in enumerate(nnco):
            merged_dict = self.sub(
                merged_dict, "NAM_OBS", "COBS_M(@VAR@)", cobs_m[obs], var=obs + 1
            )
            merged_dict = self.sub(
                merged_dict, "NAM_OBS", "NNCO(@VAR@)", obs_val, var=obs + 1
            )
            merged_dict = self.sub(
                merged_dict, "NAM_OBS", "XERROBS_M(@VAR@)", xerrobs_m[obs], var=obs + 1
            )
            if nnco[obs] == 1:
                nobstype += 1
        merged_dict = self.delete(merged_dict, "NAM_OBS", "COBS_M(@VAR@)")
        merged_dict = self.delete(merged_dict, "NAM_OBS", "NNCO(@VAR@)")
        merged_dict = self.delete(merged_dict, "NAM_OBS", "XERROBS_M(@VAR@)")
        merged_dict = self.sub(merged_dict, "NAM_OBS", "NOBSTYPE", nobstype)
        lswe = self.config.get_setting("SURFEX#ASSIM#OBS#LSWE")
        merged_dict = self.sub(merged_dict, "NAM_OBS", "LSWE", lswe)
        return merged_dict

    def set_soda_namelist(self, merged_dict):
        """Set SODA namelist.

        Args:
            merged_dict (dict): Merged dict

        Raises:
            RuntimeError: You must provide a DTG when using a list for snow
            RuntimeError: Mismatch in nncv/cvar_m
            RuntimeError: Mismatch in nncv/cvar_m/xsigma_m/xtprt_m

        Returns:
            dict: Merged dict.
        """
        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/soda.json"
        )

        merged_dict = self.sub(merged_dict, "NAM_ASSIM", "LASSIM", True)

        # Set observations
        merged_dict = self.set_obs(merged_dict)

        # LSM
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "CFILE_FORMAT_LSM",
            self.config.get_setting("SURFEX#ASSIM#CFILE_FORMAT_LSM"),
        )

        # Sea
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "CASSIM_SEA",
            self.config.get_setting("SURFEX#ASSIM#SCHEMES#SEA"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "CFILE_FORMAT_SST",
            self.config.get_setting("SURFEX#ASSIM#SEA#CFILE_FORMAT_SST"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "LREAD_SST_FROM_FILE",
            self.config.get_setting("SURFEX#ASSIM#SEA#LREAD_SST_FROM_FILE"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "LEXTRAP_SEA",
            self.config.get_setting("SURFEX#ASSIM#SEA#LEXTRAP_SEA"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "LECSST",
            self.config.get_setting("SURFEX#ASSIM#SEA#LECSST"),
        )

        # Water
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "CASSIM_WATER",
            self.config.get_setting("SURFEX#ASSIM#SCHEMES#INLAND_WATER"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "LWATERTG2",
            self.config.get_setting("SURFEX#ASSIM#INLAND_WATER#LWATERTG2"),
        )
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "LEXTRAP_WATER",
            self.config.get_setting("SURFEX#ASSIM#INLAND_WATER#LEXTRAP_WATER"),
        )

        # Nature
        merged_dict = self.sub(
            merged_dict,
            "NAM_ASSIM",
            "CASSIM_ISBA",
            self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA"),
        )

        # Snow
        laesnm = False
        snow_cycles = self.config.get_setting("SURFEX#ASSIM#ISBA#UPDATE_SNOW_CYCLES")
        if len(snow_cycles) > 0:
            logging.debug("%s", self.dtg)
            if self.dtg is not None:
                for cycle in snow_cycles:
                    logging.debug(self.dtg.strftime("%H"))
                    if int(self.dtg.strftime("%H")) == int(cycle):
                        logging.debug("true")
                        laesnm = True
            else:
                raise RuntimeError(
                    "You must provide a DTG when using a list for snow  "
                    "assimilation cycles"
                )
        merged_dict = self.sub(merged_dict, "NAM_ASSIM", "LAESNM", laesnm)

        # Set OI settings
        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "OI":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/soda_isba_oi.json"
            )
            ua_physics = self.config.get_setting("FORECAST#PHYSICS", abort=False)
            if ua_physics is None:
                logging.warning("FORECAST#PHYSICS not set. Assume arome physics")
                ua_physics = "arome"
            if ua_physics == "arome":
                merged_dict = self.sub(merged_dict, "NAM_ASSIM", "LAROME", True)
            elif ua_physics == "alaro":
                merged_dict = self.sub(merged_dict, "NAM_ASSIM", "LAROME", False)

            merged_dict = self.sub(
                merged_dict,
                "NAM_NACVEG",
                "XSIGT2MO",
                self.config.get_setting("SURFEX#ASSIM#ISBA#OI#XSIGT2MO"),
            )
            merged_dict = self.sub(
                merged_dict,
                "NAM_NACVEG",
                "XSIGH2MO",
                self.config.get_setting("SURFEX#ASSIM#ISBA#OI#XSIGH2MO"),
            )
            merged_dict = self.sub(merged_dict, "NAM_NACVEG", "XRCLIMCA", 0.0)
            merged_dict = self.sub(merged_dict, "NAM_NACVEG", "XRCLISST", 0.05)
            merged_dict = self.sub(merged_dict, "NAM_NACVEG", "NECHGU", self.fcint)
            merged_dict = self.sub(merged_dict, "NAM_NACVEG", "LOBS2M", True)
            merged_dict = self.sub(merged_dict, "NAM_NACVEG", "LOBSWG", False)
            f_clim = self.config.get_setting("SURFEX#ASSIM#ISBA#OI#CFILE_FORMAT_CLIM")
            merged_dict = self.sub(merged_dict, "NAM_ASSIM", "CFILE_FORMAT_CLIM", f_clim)
            f_fg = self.config.get_setting("SURFEX#ASSIM#ISBA#OI#CFILE_FORMAT_FG")
            merged_dict = self.sub(merged_dict, "NAM_ASSIM", "CFILE_FORMAT_FG", f_fg)

        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "EKF":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/soda_isba_ekf.json"
            )
            nvar = 0
            llincheck = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#LLINCHECK")
            merged_dict = self.sub(merged_dict, "NAM_ASSIM", "LLINCHECK", llincheck)
            xalpha = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XALPHA")
            merged_dict = self.sub(merged_dict, "NAM_VAR", "XALPHA", xalpha)
            cvar_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#CVAR_M")
            xsigma_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XSIGMA_M")
            xtprt_m = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XTPRT_M")
            nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#NNCV")
            if (
                len(nncv) != len(cvar_m)
                or len(nncv) != len(xsigma_m)
                or len(nncv) != len(xtprt_m)
            ):
                raise RuntimeError("Mismatch in nncv/cvar_m/xsigma_m/xtprt_m")
            for var, cvar_val in enumerate(cvar_m):
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "CVAR_M(@VAR@)", cvar_val, var=var + 1
                )
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "NNCV(@VAR@)", nncv[var], var=var + 1
                )
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "XTPRT_M(@VAR@)", xtprt_m[var], var=var + 1
                )
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "XSIGMA_M(@VAR@)", xsigma_m[var], var=var + 1
                )
                if nncv[var] == 1:
                    nvar += 1
            merged_dict = self.delete(merged_dict, "NAM_VAR", "XSIGMA_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "XTPRT_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "CVAR_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "NNCV(@VAR@)")
            merged_dict = self.sub(merged_dict, "NAM_VAR", "NIVAR", 0)
            merged_dict = self.sub(merged_dict, "NAM_VAR", "NVAR", nvar)
            xscale_q = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#XSCALE_Q")
            merged_dict = self.sub(merged_dict, "NAM_VAR", "XSCALE_Q", xscale_q)
            merged_dict = self.sub(merged_dict, "NAM_IO_VARASSIM", "LPRT", False)
            merged_dict = self.sub(
                merged_dict,
                "NAM_IO_VARASSIM",
                "LBEV",
                self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#EVOLVE_B"),
            )
            merged_dict = self.sub(
                merged_dict,
                "NAM_IO_VARASSIM",
                "LBFIXED",
                not self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#EVOLVE_B"),
            )

        if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "ENKF":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/soda_isba_enkf.json"
            )
            nvar = 0
            cvar_m = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#CVAR_M")
            nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#NNCV")
            if len(nncv) != len(cvar_m):
                raise RuntimeError("Mismatch in nncv/cvar_m")
            for var, cvar_val in enumerate(cvar_m):
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "CVAR_M(@VAR@)", cvar_val, var=var + 1
                )
                merged_dict = self.sub(
                    merged_dict, "NAM_VAR", "NNCV(@VAR@)", nncv[var], var=var + 1
                )

                if nncv[var] == 1:
                    nvar += 1
            merged_dict = self.delete(merged_dict, "NAM_VAR", "CVAR_M(@VAR@)")
            merged_dict = self.delete(merged_dict, "NAM_VAR", "NNCV(@VAR@)")
            merged_dict = self.sub(merged_dict, "NAM_VAR", "NIVAR", 0)
            merged_dict = self.sub(merged_dict, "NAM_VAR", "NVAR", nvar)

        # Town
        town_setting = self.config.get_setting("SURFEX#ASSIM#SCHEMES#TEB")
        merged_dict = self.sub(merged_dict, "NAM_ASSIM", "CASSIM_TEB", town_setting)
        return merged_dict

    def epilog(self, merged_dict):
        """Epilog."""
        # Always set these
        if self.config.get_setting("SURFEX#SEA#ICE") == "SICE":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/epilog_sice.json"
            )

        merged_dict = self.merge_json_namelist_file(
            merged_dict, self.input_path + "/epilog_treedrag.json"
        )
        lfaketrees = self.config.get_setting("SURFEX#TREEDRAG#FAKETREES", abort=False)
        if lfaketrees is not None:
            merged_dict = self.sub(merged_dict, "NAM_TREEDRAG", "LFAKETREE", lfaketrees)

        if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/epilog_flake.json"
            )
        return merged_dict

    def override(self, merged_dict):
        """Overide."""
        # Override posssibility
        if os.path.exists(self.input_path + "/override.json"):
            logging.warning(
                "WARNING: Override settings with content from %s/override.json",
                self.input_path,
            )
            merged_dict = self.merge_json_namelist_file(
                merged_dict, self.input_path + "/override.json"
            )
        return merged_dict

    @staticmethod
    def sub(merged_dict, nam_block, key, value, vtype=None, decade=None, var=None):
        """Substitute key with value.

        Args:
            merged_dict (dict): Merged dict
            nam_block (str): Namelist block
            key (str): Key
            value (any): Value
            vtype(int, optional): Veg type. Defaults to None.
            decade(int, optional): Decade type. Defaults to None.
            var(int, optional): Var type. Defaults to None.

        Returns:
            dict: Merged dict

        """
        nam_block = nam_block.upper()
        key = key.upper()
        if nam_block in merged_dict:
            if key in merged_dict[nam_block]:
                setting = merged_dict[nam_block][key]
                if vtype is not None:
                    if isinstance(key, str):
                        if "@VEGTYPE@" in setting:
                            key = key.replace("@VEGTYPE@", str(vtype))
                    if isinstance(setting, str):
                        if "@VEGTYPE@" in setting:
                            setting = setting.replace("@VEGTYPE@", str(vtype))
                if decade is not None:
                    if isinstance(key, str):
                        if "@DECADE@" in setting:
                            key = key.replace("@DECADE@", str(decade))
                    if isinstance(setting, str):
                        if "@DECADE@" in setting:
                            setting = setting.replace("@DECADE@", str(decade))
                if var is not None:
                    if isinstance(key, str):
                        if "@VAR@" in setting:
                            key = key.replace("@VAR@", str(var))
                    if isinstance(setting, str):
                        if "@VAR@" in setting:
                            setting = setting.replace("@VAR@", str(var))
                logging.debug("key=%s, setting=%s", key, setting)
                if isinstance(setting, str):
                    if "@" + key + "@" in setting:
                        setting = value
                    else:
                        logging.warning(
                            "No string substitution. Using existing value: %s", setting
                        )
                merged_dict[nam_block][key] = setting
            else:
                logging.warning("No setting found in namelist for: %s", key)
        else:
            logging.warning("No namelist block found in namelist for: %s", nam_block)
        return merged_dict

    @staticmethod
    def delete(merged_dict, nam_block, key):
        """Delete key from namelist block.

        Args:
            merged_dict (_type_): _description_
            nam_block (_type_): _description_
            key (_type_): _description_

        Returns:
            _type_: _description_

        """
        nam_block = nam_block.upper()
        key = key.upper()
        if nam_block in merged_dict:
            if key in merged_dict[nam_block]:
                del merged_dict[nam_block][key]
        return merged_dict

    @staticmethod
    def get_filetype_from_suffix(fname):
        """Get the file type from suffix.

        Args:
            fname (str): File name

        Returns:
            tuple: fname, format

        """
        if fname.endswith(".dir"):
            fname = fname.replace(".dir", "")
            return fname, "DIRECT"
        if fname.endswith(".ascllv"):
            fname = fname.replace(".ascllv", "")
            return fname, "ASCLLV"
        return fname, None

    @staticmethod
    def set_direct_data_namelist(lnamelist_section, ldtype, ldname, linput_path):
        """Set direct data namelist.

        Args:
            lnamelist_section (_type_): _description_
            ldtype (_type_): _description_
            ldname (_type_): _description_
            linput_path (_type_): _description_

        Returns:
            _type_: _description_

        """
        if ldname.endswith(".dir"):
            basename = os.path.splitext(os.path.basename(ldname))[0]
            filetype_name = ldtype
            if ldtype == "YSOC_TOP" or ldtype == "YSOC_SUB":
                filetype_name = "YSOC"
            return {
                "json": json.loads(
                    '{"'
                    + lnamelist_section
                    + '": { "'
                    + ldtype
                    + '": "'
                    + basename
                    + '", '
                    + '"'
                    + filetype_name
                    + 'FILETYPE": "DIRECT"}}'
                )
            }
        if ldname.endswith(".json"):
            return {"file": linput_path + "/" + ldname}

    @staticmethod
    def set_dirtyp_data_namelist(
        lnamelist_section, ldtype, ldname, vtype=None, decade=None
    ):
        """Set dirtyp data namelist.

        Args:
            lnamelist_section (_type_): _description_
            ldtype (_type_): _description_
            ldname (_type_): _description_
            vtype (_type_, optional): _description_. Defaults to None.
            decade (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_

        """
        basename = os.path.splitext(os.path.basename(ldname))[0]
        filetype_name = ldtype.upper()
        if vtype is not None or decade is not None:
            filetype_name = filetype_name + "("
        if vtype is not None:
            filetype_name = filetype_name + str(vtype)
        if vtype is not None and decade is not None:
            filetype_name = filetype_name + ","
        if decade is not None:
            filetype_name = filetype_name + str(decade)
        if vtype is not None or decade is not None:
            filetype_name = filetype_name + ")"
        return {
            "json": json.loads(
                '{"'
                + lnamelist_section
                + '": { "CFNAM_'
                + filetype_name
                + '": "'
                + basename
                + '", "CFTYP_'
                + filetype_name
                + '": "DIRTYPE"}}'
            )
        }

    @staticmethod
    def capitalize_namelist_dict(dict_in):
        """Capitalize namelist.

        Args:
            dict_in (_type_): _description_

        Returns:
            _type_: _description_

        """
        new_dict = {}
        for key in dict_in:
            upper_case2 = {}
            for key2 in dict_in[key]:
                upper_case2.update({key2.upper(): dict_in[key][key2]})
            new_dict.update({key.upper(): upper_case2})
        return new_dict

    @staticmethod
    def lower_case_namelist_dict(dict_in):
        """Lower case namelist.

        Args:
            dict_in (dict): Namelist dictionary to lower case

        Returns:
            dict: Namelist in lower case

        """
        new_dict = {}
        for key in dict_in:
            lower_case_dict = {}
            for key2 in dict_in[key]:
                lower_case_dict.update({key2.lower(): dict_in[key][key2]})
            new_dict.update({key.lower(): lower_case_dict})
        return new_dict

    @staticmethod
    def merge_namelist_dicts(old_dict, new_dict):
        """Merge namelist dicts.

        Args:
            old_dict (dict): Old dictionary
            new_dict (dict): New dictionary

        Returns:
            dict: Merged dictionary

        """
        old_dict = Namelist.capitalize_namelist_dict(old_dict)
        new_dict = Namelist.capitalize_namelist_dict(new_dict)
        merged_dict = old_dict

        for new_key, new_val in new_dict.items():
            # Namelist block already exists
            if new_key in merged_dict:
                settings = merged_dict[new_key]
                for new_key2 in new_val:
                    settings.update({new_key2: new_dict[new_key][new_key2]})

                merged_dict.update({new_key: settings})
            # New namelist block
            else:
                merged_dict.update({new_key: new_val})

        return merged_dict

    @staticmethod
    def ascii2nml(input_data):
        """Convert dict to a namelist object.

        Args:
            input_data (dict): Namelist settings

        Returns:
            f90nml.Namelist: Namelist object.
        """
        output_data = f90nml.Namelist(input_data)
        return output_data

    @staticmethod
    def ascii_file2nml(input_fname, input_fmt="json"):
        """Convert a file wih namelist settings to a Namelist object.

        Args:
            input_fname (str): Filname
            input_fmt (str, optional): File format. Defaults to "json".

        Returns:
            f90nml.Namelist: Namelist object.

        """
        if input_fmt == "json":
            with open(input_fname, mode="r", encoding="utf-8") as input_file:
                output_data = json.load(input_file)
        elif input_fmt == "yaml":
            with open(input_fname, mode="r", encoding="utf-8") as input_file:
                output_data = yaml.safe_load(input_file)
        output_data = f90nml.Namelist(output_data)
        return output_data

    @staticmethod
    def nml2ascii(input_data, output_file, output_fmt="json", indent=2):
        """Dump a namelist object as a dict in a json or yaml file.

        Args:
            input_data (f90nml.Namelist): Namelist object.
            output_file (str): Filename
            output_fmt (str, optional): File format. Defaults to "json".
            indent (int, optional): Indentation. Defaults to 2.

        """
        if output_fmt == "json":
            input_data = input_data.todict(complex_tuple=True)
            json.dump(
                input_data,
                open(output_file, "w", encoding="utf-8"),
                indent=indent,
                separators=(",", ": "),
            )
        elif output_fmt == "yaml":
            input_data = input_data.todict(complex_tuple=True)
            yaml.dump(input_data, output_file, default_flow_style=False)

    @staticmethod
    def merge_json_namelist_file(old_dict, my_file):
        """Merge json files with namelists.

        Args:
            old_dict (dict): Exististing settings
            my_file (str): Filename with new settings

        Returns:
            dict: Merged settings.

        """
        logging.debug(my_file)
        if os.path.exists(my_file):
            with open(my_file, "r", encoding="utf-8") as file_handler:
                new_dict = json.load(file_handler)
        else:
            logging.info("Possible namelist json file %s does not exist", my_file)
            new_dict = {}

        return Namelist.merge_namelist_dicts(old_dict, new_dict)

    def get_namelist(self):
        """Get namelist."""
        logging.debug("Constructing namelist:")
        merged_json_settings = self.namelist_dict
        return self.ascii2nml(merged_json_settings)



"""Obsoul."""
import logging

import numpy as np

from .datetime_utils import as_datetime, as_timedelta
from .obs import ObservationSet
from .observation import Observation


class ObservationDataSetFromObsoul(ObservationSet):
    """Observation set from obsoul file."""

    def __init__(
        self,
        content,
        an_time=None,
        label="",
        obnumber=None,
        obtypes=None,
        subtypes=None,
        neg_dt=None,
        pos_dt=None,
    ):
        """Constuct obs set from a obsoul json file.

        Args:
            content(str): Obsoul content
            an_time (as_datetime, optional): Analysis time
            neg_dt  (int, optional): Negative timedelta. Defaults to None.
            pos_dt  (int, optional): Positive timedelta. Defaults to None.
            obtypes  (list, optional): Observation types. Defaults to None.
            subtypes  (list, optional): Observation sub types. Defaults to None.
            obnumber (int, optional): Observation number. Defaults to None.
            label (str, optional): Label for observation set. Defaults to "".
        """
        observations = []
        lineno = 0
        first_line = True
        lines = content.split("\n")
        logging.debug("Found %s lines", lines)
        while lineno < len(lines) - 1:
            line = lines[lineno]
            if line.strip() == "":
                lineno += 1
            if first_line:
                line = lines[lineno]
                parts = line.split()
                logging.debug("First row parts %s", parts)
                first_line = False
                lineno += 1

            # Header
            line = lines[lineno]
            logging.debug("Header %s", line)
            obt = int(line[4:7])
            subt = int(line[7:17])
            lon = float(line[17:27])
            lat = float(line[27:38])
            stid = line[38:50].replace("'", "").strip()
            elev = np.nan
            date = line[50:60].strip()
            time = int(line[60:67].strip())
            yyyymmdhhmm = f"{date}{time:06d}"
            logging.debug("Date=%s Time=%s", date, time)
            obtime = as_datetime(yyyymmdhhmm)
            records = int(line[80:86])
            lineno += 1

            for __ in range(0, records):
                line = lines[lineno]
                parts = line.split()
                logging.debug("Data parts %s", parts)
                obn = int(parts[0])
                add_obs = True
                logging.debug("obtypes=%s obn=%s", obtypes, obn)
                if obtypes is not None:
                    if obt in obtypes:
                        if subtypes is not None:
                            if subt in subtypes:
                                pass
                            else:
                                logging.debug("Wrong subtype %s %s", subt, subtypes)
                                add_obs = False
                    else:
                        logging.debug("Wrong obtype %s %s", obt, obtypes)
                        add_obs = False

                if obnumber is not None:
                    if obn != obnumber:
                        logging.debug("Wrong obnumber %s %s", obn, obnumber)
                        add_obs = False

                # Remove if outside window
                if an_time is not None:
                    if neg_dt is not None and pos_dt is not None:
                        n_dt = as_timedelta(seconds=neg_dt)
                        p_dt = as_timedelta(seconds=pos_dt)
                        if (an_time - n_dt) <= obtime <= (an_time + p_dt):
                            pass
                        else:
                            logging.debug(
                                "Outside time window %s %s %s %s",
                                obtime,
                                an_time,
                                n_dt,
                                p_dt,
                            )
                            add_obs = False
                    else:
                        logging.debug(
                            "Not checking time window. neg_dt=%s and/or pos_dt=%s are None",
                            neg_dt,
                            pos_dt,
                        )
                value = parts[3]
                logging.debug("Obs %s %s %s %s", lineno, obn, obnumber, add_obs)
                if add_obs:
                    observations.append(
                        Observation(
                            obtime,
                            lon,
                            lat,
                            value,
                            elev=elev,
                            stid=stid,
                            varname=str(obnumber),
                        )
                    )
                lineno += 1
        logging.debug("nObs %s", len(observations))
        ObservationSet.__init__(self, observations, label=label)


class ObservationDataSetFromObsoulFile(ObservationDataSetFromObsoul):
    """Observation set from obsoul file."""

    def __init__(
        self,
        filename,
        an_time=None,
        neg_dt=None,
        pos_dt=None,
        label="",
        obnumber=None,
        obtypes=None,
        subtypes=None,
    ):
        """Constuct obs set from a obsoul file.

        Args:
            filename (str): File name
            an_time (as_datetime, optional): Analysis time
            neg_dt  (int, optional): Negative timedelta. Defaults to None.
            pos_dt  (int, optional): Positive timedelta. Defaults to None.
            obtypes  (list, optional): Observation types. Defaults to None.
            subtypes  (list, optional): Observation sub types. Defaults to None.
            obnumber (int, optional): Observation number. Defaults to None.
            label (str, optional): Label for observation set. Defaults to "".
        """
        logging.info("Opening OBSOUL file %s", filename)
        with open(filename, mode="r", encoding="utf8") as fhandler:
            content = fhandler.read()
            ObservationDataSetFromObsoul.__init__(
                self,
                content,
                an_time=an_time,
                label=label,
                obnumber=obnumber,
                pos_dt=pos_dt,
                neg_dt=neg_dt,
                obtypes=obtypes,
                subtypes=subtypes,
            )



"""Cache."""
import logging

from .datetime_utils import as_datetime_args


class Cache:
    """Cache."""

    def __init__(self, max_age):
        """Construct cache.

        Args:
            max_age (int): Maximum age in seconds.

        """
        self._files = []
        self.max_age = max_age
        self.file_handler = []
        self.interpolators = {}
        self.saved_fields = {}

    @property
    def files(self):
        """Files property."""
        return self._files

    def set_file_handler(self, filename, file_handler):
        """Set file handler.

        Args:
            filename (_type_): _description_
            file_handler (_type_): _description_

        """
        self.files.append(filename)
        self.file_handler.append(file_handler)
        logging.debug("filename -> %s", str(filename))
        logging.debug("file_handler -> %s", str(file_handler))

    def get_file_handler(self, filename):
        """Get the file handler.

        Args:
            filename (_type_): _description_

        Returns:
            _type_: _description_

        """
        f_h = None
        for findex, check_fname in enumerate(self.files):
            if check_fname == filename:
                f_h = self.file_handler[findex]
        return f_h

    def file_open(self, filename):
        """Test if file is open.

        Args:
            filename (str): Filename

        Returns:
            _type_: _description_
        """
        found = False
        for check_fname in self.files:
            if check_fname == filename:
                found = True
        return found

    def interpolator_is_set(self, inttype, geo_in, geo_out):
        """Check if interpolator is set.

        Args:
            inttype (_type_): _description_
            geo_in (_type_): _description_
            geo_out (_type_): _description_

        Returns:
            _type_: _description_

        """
        identifier_in = geo_in.identifier()
        identifier_out = geo_out.identifier()
        if inttype in self.interpolators:
            logging.debug("interpolators: %s", self.interpolators)
            logging.debug("inttype: %s", inttype)
            logging.debug("identifier_out: %s", self.interpolators[inttype])
            if identifier_out in self.interpolators[inttype]:
                logging.debug("identifier_out: %s", identifier_out)
                logging.debug(
                    "interpolators: %s", self.interpolators[inttype][identifier_out]
                )
                if identifier_in in self.interpolators[inttype][identifier_out]:
                    return True
                else:
                    logging.debug("Could not find in %s", identifier_in)
                    return False
            else:
                logging.debug("Could not find out %s", identifier_out)
                return False
        else:
            return False

    def get_interpolator(self, inttype, geo_in, geo_out):
        """Get interpolator.

        Args:
            inttype (_type_): _description_
            geo_in (_type_): _description_
            geo_out (_type_): _description_

        Returns:
            _type_: _description_

        """
        identifier_in = geo_in.identifier()
        identifier_out = geo_out.identifier()
        if self.interpolator_is_set(inttype, geo_in, geo_out):
            return self.interpolators[inttype][identifier_out][identifier_in]
        else:
            return None

    def update_interpolator(self, inttype, geo_in, geo_out, value):
        """Update interpolator.

        Args:
            inttype (_type_): _description_
            geo_in (_type_): _description_
            geo_out (_type_): _description_
            value (_type_): _description_

        """
        identifier_in = geo_in.identifier()
        identifier_out = geo_out.identifier()

        logging.debug(
            "Update interpolator %s %s %s", inttype, identifier_in, identifier_out
        )
        if inttype in self.interpolators:
            out_geos = {}
            if identifier_out in self.interpolators[inttype]:
                for out_grid in self.interpolators[inttype]:
                    in_geos = {}
                    for fint in self.interpolators[inttype][out_grid]:
                        logging.debug("Found %s for grid %s", fint, out_grid)
                        in_geos.update(
                            {fint: self.interpolators[inttype][out_grid][fint]}
                        )
                    if identifier_out == out_grid:
                        logging.debug(
                            "Update: %s for out geo %s", identifier_in, identifier_out
                        )
                        in_geos.update({identifier_in: value})
                    out_geos.update({out_grid: in_geos})
            else:
                logging.debug(
                    "Setting new: %s for out geo %s", identifier_in, identifier_out
                )
                out_geos.update({identifier_out: {identifier_in: value}})
            self.interpolators.update({inttype: out_geos})
        else:
            self.interpolators.update({inttype: {identifier_out: {identifier_in: value}}})
        logging.debug("Updated interpolator: %s", self.interpolators)

    def save_field(self, id_str, field):
        """Save field."""
        logging.debug("Saving %s", id_str)
        self.saved_fields[id_str] = field

    def clean_fields(self, this_time):
        """Clean fields.

        Args:
            this_time (_type_): _description_

        """
        logging.debug("Clean fields")
        del_keys = []
        for key in self.saved_fields:
            year = int(key[-10:-6])
            month = int(key[-6:-4])
            day = int(key[-4:-2])
            hour = int(float(key[-2:]))
            field_time = as_datetime_args(year=year, month=month, day=day, hour=hour)
            time_duration = (this_time - field_time).total_seconds()
            if time_duration > self.max_age:
                del_keys.append(key)

        for del_key in del_keys:
            del self.saved_fields[del_key]

    def is_saved(self, id_str):
        """Check if saved.

        Args:
            id_str (_type_): _description_

        Returns:
            bool: If found
        """
        logging.debug(" Check: %s", id_str)
        if len(self.saved_fields) > 0:
            logging.debug("Saved fields:")
            for key in self.saved_fields:
                logging.debug(" - %s", key)
        if id_str in self.saved_fields:
            logging.debug("Found %s", id_str)
            return True
        else:
            return False

    @staticmethod
    def generate_grib_id(gribvar, filename, validtime):
        """Generate grib id.

        Args:
            gribvar (_type_): _description_
            filename (_type_): _description_
            validtime (_type_): _description_

        Raises:
            NotImplementedError: _description_

        Returns:
            _type_: _description_

        """
        if gribvar.version == 1:
            grib_id = gribvar.generate_grib_id()
            return grib_id + f"{filename.split('/')[-1]}:{validtime.strftime('%Y%m%d%H')}"
        elif gribvar.version == 2:
            grib_id = gribvar.generate_grib_id()
            return grib_id + f"{filename.split('/')[-1]}:{validtime.strftime('%Y%m%d%H')}"
        else:
            raise NotImplementedError

    @staticmethod
    def generate_netcdf_id(var, filename, validtime):
        """Generate netcdf id.

        Args:
            var (_type_): _description_
            filename (_type_): _description_
            validtime (_type_): _description_

        Returns:
            _type_: _description_

        """
        varname = var.name
        level = str(var.level)
        member = str(var.member)
        return f"{varname}{level}{member}{filename.split('/')[-1]}{validtime.strftime('%Y%m%d%H')}"

    @staticmethod
    def generate_surfex_id(varname, patches, layers, filename, validtime):
        """Generate surfex id.

        Args:
            varname (_type_): _description_
            patches (_type_): _description_
            layers (_type_): _description_
            filename (_type_): _description_
            validtime (_type_): _description_

        Returns:
            _type_: _description_

        """
        return (
            f"{varname}{patches}{layers}{filename.split('/')[-1]}"
            f"{validtime.strftime('%Y%m%d%H')}"
        )

    @staticmethod
    def generate_obs_id(varname, filename, validtime):
        """Generate obs id.

        Args:
            varname (_type_): _description_
            filename (_type_): _description_
            validtime (_type_): _description_

        Returns:
            _type_: _description_

        """
        return f"{varname}{filename.split('/')[-1]}{validtime.strftime('%Y%m%d%H')}"

    @staticmethod
    def generate_fa_id(varname, filename, validtime):
        """Generate fa id.

        Args:
            varname (_type_): _description_
            filename (_type_): _description_
            validtime (_type_): _description_

        Returns:
            _type_: _description_

        """
        return f"{varname}{filename.split('/')[-1]}{validtime.strftime('%Y%m%d%H')}"

    @staticmethod
    def generate_id(id_type, var, filename, validtime):
        """Generate id.

        Args:
            id_type (_type_): _description_
            var (_type_): _description_
            filename (_type_): _description_
            validtime (_type_): _description_

        Raises:
            NotImplementedError: _description_

        Returns:
            _type_: _description_

        """
        if id_type == "netcdf":
            return Cache.generate_netcdf_id(var, filename, validtime)
        elif id_type == "grib1" or id_type == "grib2":
            return Cache.generate_grib_id(var, filename, validtime)
        elif id_type == "fa":
            return Cache.generate_fa_id(var, filename, validtime)
        elif id_type == "surfex":
            varname = var.varname
            patches = var.patches
            layers = var.layers
            return Cache.generate_surfex_id(varname, patches, layers, filename, validtime)
        elif id_type == "obs":
            return Cache.generate_obs_id(var, filename, validtime)
        else:
            raise NotImplementedError
